from dataclasses import dataclass, field
from pathlib import Path

from octoflow.utils.config import Config

__all__ = [
    "config",
]


@dataclass
class CacheConfig:
    path: Path = "${oc.select:resources.path}/cache"


@dataclass
class DataConfig:
    path: Path = "${oc.select:resources.path}/data"


@dataclass
class ResourcesConfig:
    path: Path = "~/.octoflow"
    # for user data
    cache: CacheConfig = field(default_factory=CacheConfig)
    # for octoflow data ???
    data: DataConfig = field(default_factory=DataConfig)


@dataclass
class LoggingConfig:
    level: str = "INFO"
    format: str = "%(asctime)s %(levelname)s %(name)s %(message)s"


@dataclass
class OctoFlowConfig:
    resources: ResourcesConfig = field(default_factory=ResourcesConfig)
    logging: LoggingConfig = field(default_factory=LoggingConfig)


config = Config(OctoFlowConfig)



from __future__ import annotations

import datetime as dt
import typing
from dataclasses import is_dataclass
from typing import Any, NamedTuple, Union

import numpy as np
import pyarrow as pa

from octoflow.data.metadata import unify_metadata

try:
    from typing import _TypedDictMeta
except ImportError:
    from typing_extensions import _TypedDictMeta  # noqa: PLC2701


__all__ = [
    "MonthDayNano",
    "from_dtype",
    "infer_type",
    "is_undefined",
    "undefined",
    "unify_types",
]


class MonthDayNano(NamedTuple):
    months: int
    days: int
    nanoseconds: int


class Undefined(pa.ExtensionType):
    def __init__(self):
        super().__init__(pa.null(), "tklearn.base.arrow.undefined")

    def __arrow_ext_serialize__(self) -> bytes:  # noqa: PLW3201
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized) -> Undefined:  # noqa: PLW3201
        return Undefined()


def undefined() -> Undefined:
    return Undefined()


UNDEFINED = undefined()

pa.register_extension_type(UNDEFINED)


def is_undefined(obj: pa.DataType) -> bool:
    return UNDEFINED.equals(obj)


def from_dataclass(cls: type) -> pa.DataType:
    """Return the PyArrow data type of a dataclass.

    Parameters
    ----------
    cls : type
        The dataclass.

    Returns
    -------
    pa.DataType
        The PyArrow data type.
    """
    fields = {}
    for field_name, field_type in typing.get_type_hints(cls).items():
        f = cls.__dataclass_fields__[field_name]
        nullable = True
        if hasattr(f, "nullable"):
            nullable = f.nullable
        fields[field_name] = pa.field(
            field_name,
            from_dtype(field_type),
            nullable,
        )
    fields = list(fields.values())
    return pa.struct(fields)


def from_typed_dict(cls: _TypedDictMeta) -> pa.DataType:
    """Return the PyArrow data type of a TypedDict.

    Parameters
    ----------
    cls : _TypedDictMeta
        The TypedDict.

    Returns
    -------
    pa.DataType
        The PyArrow data type.
    """
    fields = {}
    for field_name, field_type in typing.get_type_hints(cls).items():
        fields[field_name] = pa.field(
            field_name,
            from_dtype(field_type),
            True,
        )
    fields = list(fields.values())
    return pa.struct(fields)


def from_union(args: tuple[type, ...]) -> pa.DataType:
    unified_type = undefined()
    for arg in args:
        unified_type = unify_types(unified_type, from_dtype(arg))
    return unified_type


def from_dtype(dtype: Union[type, np.dtype, None]) -> pa.DataType:
    """Return the PyArrow data type of a provided native/NumPy data type.

    Parameters
    ----------
    dtype : type | np.dtype | None
        The native or NumPy data type.

    Returns
    -------
    pa.DataType
        The PyArrow data type.
    """
    type_args = typing.get_args(dtype)
    dtype = typing.get_origin(dtype) or dtype
    # null
    if dtype is None or dtype is type(None):
        return pa.null()
    # special types
    if dtype is Union:
        return from_union(type_args)
    # basic python data types
    if isinstance(dtype, type):
        if is_dataclass(dtype):
            return from_dataclass(dtype)
        if issubclass(dtype, _TypedDictMeta):
            return from_typed_dict(dtype)
        if issubclass(dtype, list):
            (item_arg,) = type_args
            return pa.list_(pa.field("item", from_dtype(item_arg)))
        if issubclass(dtype, dict):
            key_arg, value_arg = type_args
            return pa.map_(from_dtype(key_arg), from_dtype(value_arg))
        if issubclass(dtype, str):
            return pa.string()
        if issubclass(dtype, bytes):
            return pa.binary()
        if issubclass(dtype, dt.datetime):
            # The smallest possible difference between
            #   non-equal datetime objects is timedelta(microseconds=1).
            return pa.timestamp("us", tz=None)
        if issubclass(dtype, dt.date):
            return pa.date32()
        if issubclass(dtype, dt.time):
            # The smallest possible difference between
            #   non-equal time objects, timedelta(microseconds=1).
            return pa.time64("us")
        if issubclass(dtype, dt.timedelta):
            # The smallest possible difference between
            #   non-equal timedelta objects, timedelta(microseconds=1).
            return pa.duration("us")
        if issubclass(dtype, MonthDayNano):
            return pa.month_day_nano_interval()
    # numpy data types
    if not isinstance(dtype, np.dtype):
        # convert dtype (obj/type) to np.dtype
        dtype = np.dtype(dtype)
    try:
        return pa.from_numpy_dtype(dtype)
    except NotImplementedError as ex:
        msg = f"unsupported data type '{dtype}'"
        raise ValueError(msg) from ex


def _is_compatible(
    op: callable, left: pa.DataType, right: pa.DataType
) -> bool:
    """Return whether both left and right are of the same type.

    Parameters
    ----------
    op : callable
        The function to check the type.
    left : pa.DataType
        The left PyArrow data type.
    right : pa.DataType
        The right PyArrow data type.

    Returns
    -------
    bool
        Whether both left and right are of the same type.
    """
    is_left, is_right = op(left), op(right)
    if is_left and is_right:
        return True
    if is_left or is_right:
        msg = f"types {left} and {right} are not compatible"
        raise ValueError(msg)
    return False


def unify_types(left: pa.DataType, right: pa.DataType) -> pa.DataType:
    """Return the PyArrow data type that can represent both left and right.

    Parameters
    ----------
    left : pa.DataType
        The left PyArrow data type.
    right : pa.DataType
        The right PyArrow data type.

    Returns
    -------
    pa.DataType
        The PyArrow data type.
    """
    if is_undefined(left) or pa.types.is_null(left):
        return right
    if is_undefined(right) or pa.types.is_null(right):
        return left
    if left.equals(right):
        return left
    if _is_compatible(pa.types.is_list, left, right):
        left_field, right_field = left.field(0), right.field(0)
        nullable = (
            left_field.nullable
            or left.value_type.equals(pa.null())
            or right_field.nullable
            or right.value_type.equals(pa.null())
        )
        metadata = unify_metadata(left_field, right_field)
        return pa.list_(
            pa.field(
                "item",
                unify_types(left.value_type, right.value_type),
                nullable,
                metadata,
            )
        )
    if _is_compatible(pa.types.is_struct, left, right):
        fields = {}
        # there may be fields that are in left but not in right
        #   those fields are optional i.e., nullable fields
        left_only_fields = set()
        for left_field in left:
            fields[left_field.name] = left_field
            left_only_fields.add(left_field.name)
        for right_field in right:
            right_field_name = right_field.name
            left_only_fields.discard(right_field_name)
            if right_field_name in fields:
                # existing field - promote the type
                left_field = fields[right_field_name]
                nullable = (
                    left_field.nullable
                    or left_field.type.equals(pa.null())
                    or right_field.nullable
                    or right_field.type.equals(pa.null())
                )
                try:
                    promoted_type = unify_types(
                        left_field.type, right_field.type
                    )
                except ValueError as ex:
                    msg = (
                        f"cannot unify types '{left_field.type}' and "
                        f"'{right_field.type}' of field '{right_field_name}'"
                    )
                    raise ValueError(msg) from ex
                metadata = unify_metadata(left_field, right_field)
                fields[right_field_name] = pa.field(
                    right_field_name, promoted_type, nullable, metadata
                )
            else:
                # new field (right only) is nullable
                fields[right_field_name] = right_field.with_nullable(True)
        # replace the left only fields with nullable fields
        for left_field_name in left_only_fields:
            left_field = fields[left_field_name]
            if left_field.nullable:
                # already nullable
                continue
            fields[left_field_name] = left_field.with_nullable(True)
        fields = list(fields.values())
        return pa.struct(fields)
    # for any other types, use the NumPy data type
    dtype = np.promote_types(left.to_pandas_dtype(), right.to_pandas_dtype())
    try:
        return from_dtype(dtype)
    except ValueError as ex:
        msg = f"cannot unify types '{left}' and '{right}'"
        raise ValueError(msg) from ex


def infer_type(obj: Any) -> pa.DataType:
    """Return the PyArrow data type of an object.

    Parameters
    ----------
    obj : Any
        The object.

    Returns
    -------
    pa.DataType
        The PyArrow data type.
    """
    try:
        if not np.isscalar(obj):
            # e.g., list, dict, etc.
            raise NotImplementedError
        dtype = np.min_scalar_type(obj)
        return pa.from_numpy_dtype(dtype)
    except NotImplementedError:
        pass
    if isinstance(obj, list):
        item_type = undefined()
        nullable = False
        for item in obj:
            current_item_type = infer_type(item)
            if not nullable and current_item_type.equals(pa.null()):
                nullable = True
            if is_undefined(item_type):
                item_type = current_item_type
            else:
                item_type = unify_types(item_type, current_item_type)
        value_field = pa.field("item", item_type, nullable)
        return pa.list_(value_field)
    if isinstance(obj, dict):
        fields = {}
        for key, value in obj.items():
            value_type = infer_type(value)
            nullable = value_type.equals(pa.null())
            value_field = pa.field(key, value_type, nullable)
            fields[key] = value_field
        fields = list(fields.values())
        return pa.struct(fields)
    if isinstance(obj, dt.time):
        unit = "us" if obj.microsecond else "s"
        tz = obj.tzname()
        if tz is not None:
            return pa.timestamp(unit, tz=tz)
        if unit == "s":
            return pa.time32(unit)
        return pa.time64(unit)
    return from_dtype(type(obj))



from __future__ import annotations

import weakref
from contextlib import contextmanager, suppress
from pathlib import Path
from typing import Generator, Mapping, Optional, Set, Union

from git import Repo

from octoflow import logging
from octoflow.tracking import (
    Run,
    SQLAlchemyTrackingStore,
    TrackingClient,
)
from octoflow.utils.rsync import rsync

logger = logging.get_logger(__name__)


class ProjectExperiment:
    def __init__(self, project: Project, expr_name: str) -> None:
        self.get_project = weakref.ref(project)
        self.expr_name = expr_name

    @property
    def project(self) -> Project:
        return self.get_project()

    def start_run(
        self,
        force: bool = False,
        description: Optional[str] = None,
    ) -> Run:
        commit_hash = self.project.sync()
        tracking_uri_path = (
            self.project.base_path
            / "experiments"
            / self.expr_name
            / f"{commit_hash}.db"
        )
        if tracking_uri_path.exists():
            if not force:
                msg = f"experiment {self.expr_name} has already been run"
                raise FileExistsError(msg)
            tracking_uri_path.unlink()
        # create the parent directories
        tracking_uri_path.parent.mkdir(parents=True, exist_ok=True)
        tracking_uri = f"sqlite:///{tracking_uri_path}"
        store = SQLAlchemyTrackingStore(tracking_uri)
        client = TrackingClient(store)
        expr = client.get_or_create_experiment(self.expr_name)
        return expr.start_run(commit_hash, description=description)


class ProjectExperimentDict(Mapping[str, ProjectExperiment]):
    def __init__(self, project: Project) -> None:
        self.get_project = weakref.ref(project)

    @property
    def project(self) -> Project:
        return self.get_project()

    @property
    def experiments_path(self) -> Path:
        exprs_dir = self.project.base_path / "experiments"
        exprs_dir.mkdir(exist_ok=True)
        return exprs_dir

    @property
    def names(self) -> Set[str]:
        if self.experiments_path.exists():
            set(self.experiments_path.iterdir())
        return set()

    def __iter__(self):
        yield from self.names

    def __getitem__(self, key: str) -> ProjectExperiment:
        return ProjectExperiment(self.project, key)

    def __contains__(self, key: str) -> bool:
        return key in self.names

    def __len__(self) -> int:
        return len(self.names)

    def __repr__(self) -> str:
        return f"ExperimentDict({self.project.base_path})"

    def first(self) -> ProjectExperiment:
        try:
            return next(iter(self.values()))
        except StopIteration:
            msg = "no experiments found"
            raise KeyError(msg) from None


def update_project_gitgnore(path: Path) -> None:
    gitignore_path = path / ".gitignore"
    if not gitignore_path.exists():
        gitignore_path.write_text("# octoflow\n.octoflow\n")
    else:
        gitignore_text = gitignore_path.read_text().rstrip()
        if "# octoflow" in gitignore_text:
            return
        gitignore_text += "\n# octoflow\n.octoflow\n"
        gitignore_path.write_text(gitignore_text.strip())


class Project:
    def __init__(self, path: Union[str, Path]) -> None:
        path = Path(path)
        self.name = path.name
        self.base_path = path / ".octoflow"
        # update the gitignore file
        update_project_gitgnore(path)
        self.base_path.mkdir(parents=True, exist_ok=True)
        self.sync()

    @contextmanager
    def get_repo(self) -> Generator[Repo, None, None]:
        # Initialize the git repository
        repo = Repo.init(self.base_path / "project")
        # commit the initial changes to main if there are no commits
        has_no_commits = True
        with suppress(StopIteration, ValueError):
            next(repo.iter_commits())
            has_no_commits = False
        if has_no_commits:
            repo.index.add("*")
            repo.index.commit("Initialize project")
        # close the repo to avoid memory leaks
        yield repo
        repo.close()

    def sync(self, message: Optional[str] = None) -> str:
        # use rsync to copy the project structure
        for cout in rsync(
            self.base_path.parent,
            self.base_path / "project",
            exclude=[
                ".git",
                ".gitignore",
                ".octoflow",
            ],
            append_dir=False,
        ):
            logger.info(cout)
        with self.get_repo() as repo:
            # Commit the changes to the experiment branch
            if "nothing to commit" in repo.git.status():
                commit_hash = repo.git.rev_parse("HEAD")
            else:
                repo.index.add("*")
                if message is None:
                    message = "Update project"
                repo.index.commit(message)
            commit_hash = repo.git.rev_parse("HEAD")
        return commit_hash

    @property
    def experiments(self):
        return ProjectExperimentDict(self)



import functools
import importlib

__all__ = [
    "create_object",
]


def import_class(cls: str):
    if isinstance(cls, str):
        module, class_name = cls.rsplit(".", 1)
        module = importlib.import_module(name=module)
        cls = getattr(module, class_name)
    return cls


def create_object(
    __cls: str,  # like {module}.{class_name} or {class_name}
    __partial: bool = False,
    *args,
    **kwargs,
):
    cls = import_class(__cls)
    try:
        if __partial:
            return functools.partial(cls, *args, **kwargs)
        return cls(cls, *args, **kwargs)
    except Exception as ex:
        raise ex



from __future__ import annotations

import functools
import inspect
import pickle  # noqa: S403
from typing import (
    Any,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    TypeVar,
    Union,
    runtime_checkable,
)

import dill  # noqa: S403
import xxhash

__all__ = [
    "UpdateHashFunc",
    "hash",
    "hashable",
]

T = TypeVar("T")


@runtime_checkable
class Hashable(Protocol):
    def _update_hash(self, m: xxhash.xxh64):
        raise NotImplementedError


def _update_hash_header(
    m: xxhash.xxh64,
    obj: Any,
    *,
    name: Optional[str] = None,
    deep: Optional[bool] = None,
) -> None:
    if name is None:
        module = obj.__class__.__module__
        if module is None or module == str.__class__.__module__:
            name = obj.__class__.__name__
        else:
            name = module + "." + obj.__class__.__name__
    m.update(f"==Type:{name}==".encode())


def _update_hash_bytes(
    m: xxhash.xxh64,
    obj: bytes,
    *,
    name: Optional[str] = None,
    deep: Optional[bool] = None,
) -> None:
    _update_hash_header(m, obj, name=name)
    m.update(obj)


def _update_hash_mapping(
    m: xxhash.xxh64,
    obj: Mapping,
    *,
    name: Optional[str] = None,
    deep: bool = False,
) -> None:
    _update_hash_header(m, obj, name=name)
    for key, value in obj.items():
        _update_hash(
            m,
            (key, value),
            name="tklearn.utils.hash.MappingItem",
            deep=deep,
        )


def _update_hash_sequence(
    m: xxhash.xxh64,
    obj: Sequence,
    *,
    name: Optional[str] = None,
    deep: bool = False,
) -> None:
    _update_hash_header(m, obj, name=name)
    for item in obj:
        _update_hash(
            m,
            item,
            name="tklearn.utils.hash.SequenceItem",
            deep=deep,
        )


def _update_hash_partial(
    m: xxhash.xxh64,
    obj: functools.partial,
    *,
    name: Optional[str] = None,
    deep: bool = False,
) -> None:
    _update_hash_header(m, obj, name=name)
    _update_hash(m, obj.func)
    _update_hash_sequence(
        m, obj.args, name="tklearn.utils.hash.PartialArgs", deep=deep
    )
    _update_hash_mapping(
        m,
        obj.keywords,
        name="tklearn.utils.hash.PartialKeywords",
        deep=deep,
    )


def _update_hash_any(
    m: xxhash.xxh64,
    obj: Any,
    *,
    name: Optional[str] = None,
    deep: Optional[bool] = None,
) -> None:
    _update_hash_header(m, obj, name=name)
    m.update(dill.dumps(obj))


def _update_hash(
    m: xxhash.xxh64,
    obj: Any,
    *,
    name: Optional[str] = None,
    deep: bool = False,
) -> None:
    """Update the hash with the object.

    Parameters
    ----------
    m : xxhash.xxh64
        The hash object.
    obj : Any
        The object to hash.
    name : str, optional
        The name of the object, by default None.
    deep : bool, optional
        Whether to hash the object recursively, by default False.
    """
    if isinstance(obj, bytes):
        _update_hash_bytes(m, obj, name=name)
    elif isinstance(obj, Hashable):
        obj._update_hash(m)
    elif isinstance(obj, functools.partial):
        _update_hash_partial(m, obj, name=name, deep=deep)
    elif deep and isinstance(obj, Mapping):
        _update_hash_mapping(m, obj, name=name, deep=deep)
    elif deep and isinstance(obj, Sequence) and not isinstance(obj, str):
        _update_hash_sequence(m, obj, name=name, deep=deep)
    else:
        _update_hash_any(m, obj, name=name)


def update_hash(m: xxhash.xxh64, obj: Any, deep: bool = False) -> None:
    """Update the hash with the object.

    Parameters
    ----------
    m : xxhash.xxh64
        The hash object.
    obj : Any
        The object to hash.
    deep : bool, optional
        Whether to hash the object recursively, by default False.
    """
    _update_hash(m, obj, deep=deep)


def hash(value: Any, deep: bool = False) -> str:
    """Hash the object.

    Parameters
    ----------
    value : Any
        The object to hash.
    deep : bool, optional
        Whether to hash the object recursively, by default False.

    Returns
    -------
    str
        The hash value.
    """
    m = xxhash.xxh64()
    update_hash(m, value, deep=deep)
    return m.hexdigest()


class UpdateHashFunc:
    def __init__(
        self,
        obj: Any,
        *,
        mode: str,
        version: Optional[str] = None,
    ) -> None:
        self.obj = obj
        self.mode = mode
        self.version = version

    def _update_hash_header(self, m: xxhash.xxh64) -> None:
        _update_hash_header(m, self.obj, name=None)
        # add version to hash
        version = "" if self.version is None else self.version
        m.update(f"==Version:{version}==".encode())

    def __call__(self, m: xxhash.xxh64) -> None:
        self._update_hash_header(m)
        if self.mode == "dill":
            repr_ = dill.dumps(self.obj)
        elif self.mode in {"pickle", "pkl"}:
            # works for pickleable objects
            repr_ = pickle.dumps(self.obj)
        elif self.mode in {"source", "src"}:
            # this works for functions and classes
            repr_ = inspect.getsource(self.obj).encode("utf-8")
        elif callable(self.mode):
            repr_ = self.mode(self.obj)
        else:
            msg = f"invalid mode: {self.mode}"
            raise ValueError(msg)
        m.update(repr_)


def hashable(
    mode: Union[
        Literal["dill", "pickle", "pkl", "source", "src"],
        callable,
    ] = "dill",
    version: None[str] = None,
    obj: Optional[T] = None,
) -> T:
    if obj is None:
        return functools.partial(hashable, mode, version)
    obj._update_hash = UpdateHashFunc(obj, mode=mode, version=version)
    return obj



from __future__ import annotations

from typing import Final, Generic, TypeVar

import pyarrow.dataset as ds
from typing_extensions import ParamSpec

__all__ = [
    "DEFAULT_BATCH_SIZE",
    "DEFAULT_FORMAT",
    "BaseDataset",
    "BaseDatasetLoader",
    "BaseExpression",
    "PyArrowWrapper",
]

ArrowType = TypeVar("ArrowType")

P = ParamSpec("P")

R = TypeVar("R")

DEFAULT_BATCH_SIZE: Final[int] = 1_048_576

DEFAULT_FORMAT: Final[str] = "arrow"


class PyArrowWrapper(Generic[ArrowType]):
    def __init__(self, wrapped: ArrowType) -> None:
        self._wrapped_ = wrapped

    @property
    def _wrapped(self) -> ArrowType:
        return self._wrapped_

    @_wrapped.setter
    def _wrapped(self, value: ArrowType) -> None:
        self._wrapped_ = value

    def to_pyarrow(self) -> ArrowType:
        return self._wrapped


BaseExpression = PyArrowWrapper[ds.Expression]

BaseDataset = PyArrowWrapper[ds.Dataset]


class BaseDatasetLoader(Generic[P, R]): ...



import re

__all__ = [
    "escape",
    "unescape",
]

ESCAPED_CHARS = r"\\`*_{}[]()#+\-.!"
ESCAPED_CHARS_RE = re.compile(rf"([{re.escape(ESCAPED_CHARS)}])")
UNESCAPED_CHARS_RE = re.compile(rf"\\([{re.escape(ESCAPED_CHARS)}])")


def escape(text: str) -> str:
    """Escape text."""
    return ESCAPED_CHARS_RE.sub(r"\\\1", text)


def unescape(text: str) -> str:
    """Unescape text."""
    return UNESCAPED_CHARS_RE.sub(r"\1", text)



from typing import Any, Optional

__all__ = [
    "unify_metadata",
]


def unify_metadata(left: Any, right: Any) -> Optional[dict]:
    # merge metadata left then right
    #   (right metadata will overwrite left metadata)
    metadata = left.metadata
    if metadata is None:
        metadata = right.metadata
    elif right.metadata is not None:
        # metadata and right.metadata are not None
        metadata = {**metadata, **right.metadata}
    return metadata



import shutil
import tempfile
from contextlib import suppress
from pathlib import Path
from typing import Union

from octoflow import config

__all__ = [
    "cache",
]


class Cache:
    def __init__(self, path: Union[str, Path, None] = None) -> None:
        if isinstance(path, str):
            path = Path(path).expanduser()
        if path is None:
            with suppress(AttributeError):
                path = config.resources.cache.path
        if path is None:
            with suppress(AttributeError):
                path = Path(config.resources.path).expanduser() / "cache"
        if path is None:
            path = Path(tempfile.gettempdir()) / "octoflow" / "cache"
        self._path = path

    @property
    def path(self) -> Path:
        return self._path

    def clean(self):
        if self.path.exists():
            shutil.rmtree(self.path)


cache = Cache()



from dataclasses import _MISSING_TYPE, MISSING

__all__ = [
    "MISSING",
    "MissingType",
    "DefaultType",
    "DEFAULT",
]

MissingType = _MISSING_TYPE


class DefaultType:
    pass


DEFAULT = DefaultType()



import shlex
import shutil
import time
from pathlib import Path
from subprocess import PIPE, Popen  # noqa: S404
from tempfile import NamedTemporaryFile
from typing import Generator, Optional

__all__ = [
    "rsync",
]


class RSyncError(Exception):
    pass


def rsync(
    src: Path,
    dest: Path,
    exclude: Optional[list[str]] = None,
    ignore_errors: Optional[bool] = False,
    append_dir: Optional[bool] = False,
) -> Generator[str, None, None]:
    src = Path(src) if isinstance(src, str) else src
    dest = Path(dest) if isinstance(dest, str) else dest
    exclude = [] if exclude is None else exclude
    exclude = " ".join([f"--exclude={e}" for e in exclude])
    rsync_path = shutil.which("rsync")
    src = src if append_dir else f"{src}/"
    shell_cmd = f"{rsync_path} -av {exclude} {src} {dest}"
    shell_cmd_parts = shlex.split(shell_cmd)
    with NamedTemporaryFile() as out, NamedTemporaryFile() as err:
        proc = Popen(
            shell_cmd_parts,  # noqa: S603
            stdout=out,
            stderr=err,
            stdin=PIPE,
        )
        # block until the process is finished
        stdout, stderr, stdout_lines = "", "", []
        with open(out.name, encoding="utf-8") as fout, open(
            err.name, encoding="utf-8"
        ) as ferr:
            while proc.poll() is None:
                stdout += fout.read()
                stdout_lines_current = stdout.split("\n")
                stdout_lines += stdout_lines_current[:-1]
                stdout = stdout_lines_current[-1]
                stderr += ferr.read()
                if len(stdout_lines_current) > 0:
                    yield from stdout_lines_current[:-1]
                time.sleep(0.01)
            stdout += fout.read()
            stdout_lines_current = stdout.split("\n")
            yield from stdout_lines_current
            stderr += ferr.read()
        if len(stderr.strip()) != 0 and not ignore_errors:
            raise RSyncError(stderr)



import functools
import itertools
from contextlib import suppress
from typing import Any, Dict, Generator, Optional, Tuple, TypeVar

import pyarrow as pa
from typing_extensions import Self

from octoflow.data import types
from octoflow.data.metadata import unify_metadata
from octoflow.data.types import infer_type, unify_types
from octoflow.exceptions import ValidationError

__all__ = [
    "infer_schema",
    "unify_schemas",
]

T = TypeVar("T")


def unify_schemas(this: pa.Schema, other: Optional[pa.Schema]) -> pa.Schema:
    if other is None:
        return this
    _fields = {}
    this_schema_fields = set()
    for this_field in this:
        this_field_name = this_field.name
        _fields[this_field.name] = pa.field(
            this_field.name,
            this_field.type,
            this_field.nullable,
            this_field.metadata,
        )
        this_schema_fields.add(this_field_name)
    for other_field in other:
        other_field_name = other_field.name
        this_schema_fields.discard(other_field_name)
        if other_field_name in _fields:
            this_field = _fields[other_field_name]
            try:
                promoted_type = unify_types(this_field.type, other_field.type)
            except ValueError as ex:
                msg = (
                    f"cannot unify types '{this_field.type}' and "
                    f"'{other_field.type}' of field '{other_field_name}'"
                )
                raise ValueError(msg) from ex
            nullable = (
                this_field.nullable
                or this_field.type.equals(pa.null())
                or other_field.nullable
                or other_field.type.equals(pa.null())
            )
            metadata = unify_metadata(this_field, other_field)
            _fields[other_field_name] = pa.field(
                other_field_name,
                promoted_type,
                nullable,
                metadata,
            )
        else:
            _fields[other_field_name] = pa.field(
                other_field_name,
                other_field.type,
                True,  # nullable because this field is not in the existing
                other_field.metadata,
            )
    for this_field_name in this_schema_fields:
        this_field = _fields[this_field_name]
        _fields[this_field_name] = pa.field(
            this_field_name,
            this_field.type,
            True,  # nullable because this field is not in the other schema
            this_field.metadata,
        )
    metadata = unify_metadata(this, other)
    return pa.schema(list(_fields.values()), metadata=metadata)


def infer_schema(
    data: Dict[str, Any],
    metadata: Optional[Dict[str, Any]] = None,
) -> Self:
    _fields = {}
    for key, value in data.items():
        dtype = infer_type(value)
        _fields[key] = pa.field(key, dtype, True, None)
    return pa.schema(list(_fields.values()), metadata=metadata)


def validate(schema: pa.Schema, data: dict) -> bool:
    """
    Validates a dictionary against a PyArrow schema.

    Parameters
    ----------
    schema : pyarrow.Schema
        The PyArrow schema to validate against.
    data : dict
        The dictionary to validate.

    Raises
    ------
    ValidationError
        If the dictionary does not match the schema.

    Examples
    --------
    >>> schema = pa.schema([pa.field('id', pa.int64()), pa.field('name', pa.string())])
    >>> valid_dict = {'id': 1, 'name': 'Alice'}
    >>> validate(schema, valid_dict)
    >>> invalid_dict = {'id': '1', 'name': 'Alice'}
    >>> validate(schema, invalid_dict)
    Traceback (most recent call last):
    ...
    ValidationError: ...
    """  # noqa: E501
    try:
        pa.RecordBatch.from_pylist([data], schema)
    except pa.lib.ArrowInvalid as e:
        raise ValidationError(str(e).lower()) from e


def get_schema(data: T) -> Tuple[T, pa.Schema]:
    """
    Extracts the schema from a PyArrow schema or a generator of PyArrow
    record batches.

    Parameters
    ----------
    data : Any
        The PyArrow schema or generator of record batches.

    Returns
    -------
    Tuple[Any, pa.Schema]
        The data and the schema.
    """
    with suppress(AttributeError):
        return data, data.schema
    if isinstance(data, Generator):
        data, _data = itertools.tee(data)
        try:
            return data, next(_data).schema
        except StopIteration:
            return data, pa.schema([])
    msg = (
        "expected data to be of type 'Generator', "
        f"got '{data.__class__.__name__}'"
    )
    raise TypeError(msg)


def from_dataclass(cls: T) -> pa.Schema:
    """
    Converts a dataclass to a PyArrow schema.

    Parameters
    ----------
    cls : Type[T]
        The dataclass to convert.

    Returns
    -------
    pa.Schema
        The PyArrow schema.

    Examples
    --------
    >>> import dataclasses
    >>> @dataclasses.dataclass
    ... class Record:
    ...     id: int
    ...     name: str
    >>> from_dataclass(Record)
    pyarrow.Schema([...])
    """
    fields = []
    for field in cls.__dataclass_fields__.values():
        nullable = True
        if hasattr(field, "nullable"):
            nullable = field.nullable
        fields.append(
            pa.field(
                field.name,
                types.from_dtype(field.type),
                nullable,
                None,
            )
        )
    return pa.schema(fields)


@functools.wraps(from_dataclass)
def get_schema_from_dataclass(*args, **kwargs) -> pa.Schema:
    """
    Alias for `from_dataclass`.

    Examples
    --------
    >>> import dataclasses
    >>> @dataclasses.dataclass
    ... class Record:
    ...     id: int
    ...     name: str
    >>> get_schema_from_dataclass(Record)
    pyarrow.Schema([...])
    """
    return from_dataclass(*args, **kwargs)



from __future__ import annotations

import copy
import datetime as dt
import os
from collections import UserDict, defaultdict
from dataclasses import field
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    Union,
)

from typing_extensions import Self

from octoflow.tracking import store
from octoflow.tracking.store import (
    StoredModel,
    TrackingStore,
    ValueType,
    VariableType,
)
from octoflow.utils.collections import flatten

__all__ = [
    "Experiment",
    "Run",
    "TrackingStore",
    "Value",
    "Variable",
]

JSONType = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]


class TrackingClient:
    def __init__(self, store: TrackingStore) -> None:
        super().__init__()
        self._store: Optional[TrackingStore] = store

    @property
    def store(self) -> TrackingStore:
        if self._store is None:
            msg = "store not set"
            raise RuntimeError(msg)
        return self._store

    def create_experiment(
        self,
        name: str,
        description: Optional[str] = None,
        artifact_uri: Optional[str] = None,
    ) -> Experiment:
        return self._store.create_experiment(name, description, artifact_uri)

    def get_experiment_by_name(self, name: str) -> Optional[Experiment]:
        return self._store.get_experiment_by_name(name)

    def get_or_create_experiment(
        self,
        name: str,
        *,
        description: Optional[str] = None,
        artifact_uri: Optional[str] = None,
    ) -> Experiment:
        err = None
        try:
            # try to create if exists - if fails then it is there
            return self.create_experiment(
                name,
                description,
                artifact_uri,
            )
        except ValueError as e:
            err = e
        try:
            return self.get_experiment_by_name(name)
        except ValueError as ex:
            if err is not None:
                raise err from ex
            raise ex

    def list_experiments(self):
        return self._store.list_experiments()


class Experiment(StoredModel):
    id: int = field(init=False)
    name: str
    description: Optional[str]
    artifact_uri: Optional[str]

    @store.wrap
    def start_run(self, name: str, description: Optional[str] = None) -> Run:
        return self.store.create_run(self.id, name, description)

    @store.wrap
    def search_runs(self, **kwargs) -> List[Run]:
        return self.store.search_runs(self.id, **kwargs)

    @store.wrap
    def delete_run(self, run: Union[Run, int]) -> None:
        run_id = run.id if isinstance(run, Run) else run
        self.store.delete_run(self.id, run_id)


class TagsMapping(MutableMapping[str, JSONType]):
    def __init__(self, run: Run) -> None:
        self._run = run

    def __getitem__(self, key: str) -> JSONType:
        with self._run._store:
            return self._run.store.get_tag(self._run.id, key)

    def __setitem__(self, key: str, value: JSONType) -> None:
        with self._run._store:
            self._run.store.set_tag(self._run.id, key, value)

    def __delitem__(self, key: str) -> None:
        with self._run._store:
            self._run.store.delete_tag(self._run.id, key)

    @property
    def data(self) -> Dict[str, JSONType]:
        with self._run._store:
            return self._run.store.get_tags(self._run.id)

    def __iter__(self) -> Iterator[str]:
        for key, _ in self.data.items():
            yield key

    def __len__(self) -> int:
        with self._run._store:
            return self._run.store.count_tags(self._run.id)

    def __repr__(self) -> str:
        return repr(self.data)


class Run(StoredModel):
    id: int = field(init=False)
    experiment_id: int
    name: str
    description: Optional[str]
    created_at: Optional[dt.datetime] = None

    tags: MutableMapping[str, JSONType] = field(init=False)

    def __post_init__(self):
        super().__post_init__()
        self.tags = TagsMapping(self)

    @store.wrap
    def log_param(
        self,
        key: str,
        value: ValueType,
        *,
        step: Union[Value, int, None] = None,
    ) -> Value:
        step_id = step.id if isinstance(step, Value) else step
        return self.store.log_value(
            self.id,
            key,
            value,
            step_id=step_id,
            type="param",
            is_step=None,
        )

    @store.wrap
    def log_params(
        self,
        values: Mapping[str, ValueType],
        *,
        step: Optional[Value] = None,
        prefix: Optional[str] = None,
    ) -> List[Value]:
        step_id = step.id if isinstance(step, Value) else step
        input_vals = []
        for key, value in flatten(values, parent_key=prefix).items():
            input_vals.append({
                "key": key,
                "value": value,
                "type": "param",
                "step_id": step_id,
                "is_step": None,
            })
        return self.store.log_values(self.id, input_vals)

    @store.wrap
    def log_metric(
        self,
        key: str,
        value: ValueType,
        *,
        step: Union[Value, int, None] = None,
    ) -> Value:
        step_id = step.id if isinstance(step, Value) else step
        return self.store.log_value(
            self.id,
            key,
            value,
            step_id=step_id,
            type="metric",
            is_step=False,
        )

    @store.wrap
    def log_metrics(
        self,
        values: Mapping[str, ValueType],
        *,
        step: Optional[Value] = None,
        prefix: Optional[str] = None,
    ) -> List[Value]:
        step_id = step.id if isinstance(step, Value) else step
        input_vals = []
        for key, value in flatten(values, parent_key=prefix).items():
            input_vals.append({
                "key": key,
                "value": value,
                "type": "metric",
                "step_id": step_id,
                "is_step": False,
            })
        return self.store.log_values(self.id, input_vals)

    @store.wrap
    def get_values(self) -> List[Tuple[Variable, Value]]:
        return self.store.get_values(self.id)


class Variable(StoredModel):
    id: int = field(init=False)
    experiment_id: int
    key: str
    parent_id: Optional[int]
    type: Optional[VariableType] = None
    is_step: Optional[bool] = None


class Value(StoredModel):
    id: int = field(init=False)
    run_id: int
    variable_id: int
    value: ValueType
    timestamp: Optional[dt.datetime] = None
    step_id: Optional[int] = None


class RunTags(StoredModel):
    id: int = field(init=False)
    run_id: int
    tag_id: int
    value: JSONType = None


class Tag(StoredModel):
    id: int = field(init=False)
    name: str


class TreeNode(UserDict):
    is_nested: bool = False

    @classmethod
    def _from_values(
        cls,
        root: Dict[int, Any],
        nodes: Dict[int, Tuple[str, Any]],
        /,
        path="/",
    ) -> Self:
        tree = cls()
        for node_id, inner in root.items():
            key, value = nodes[node_id]
            key_path = os.path.join(path, str(key))
            value_path = os.path.join(key_path, str(value))
            if inner is None:
                if key in tree:
                    msg = f"key path '{key_path}' already exists"
                    raise ValueError(msg)
                tree[key] = value
            elif key in tree:
                temp = tree[key]
                if not isinstance(temp, TreeNode) or not temp.is_nested:
                    msg = (
                        f"expected nested 'TreeNode' for key path '{key_path}'"
                    )
                    msg += f", got '{type(temp).__name__}'"
                    raise ValueError(msg)
                temp.update({
                    value: cls._from_values(inner, nodes, path=value_path),
                })
            else:
                subtree = TreeNode()
                subtree.is_nested = True
                subtree.update({
                    value: cls._from_values(inner, nodes, path=value_path),
                })
                tree[key] = subtree
        return tree

    @classmethod
    def from_values(cls, values: List[Tuple[Variable, Value]]) -> Self:
        nodes = {}
        is_step = {}
        for var, value in values:
            is_step[value.id] = var.is_step
            nodes[value.id] = (var.key, value.value)
        tree = {}
        for _, value in values:
            sn, vn = value.step_id, value.id
            if sn is None:
                sn = "__root__"
            if vn not in tree:
                tree[vn] = {} if is_step[value.id] else None
            if sn not in tree:
                tree[sn] = {}
            tree[sn][vn] = tree[vn]
        root = tree["__root__"]
        return cls._from_values(root, nodes)

    def _flatten(
        self,
        *,
        base_dict: Optional[dict] = None,
        ancestor_keys: Optional[Tuple[str]] = None,
    ) -> Dict[Tuple, List]:
        if base_dict is None:
            base_dict = {}
        if ancestor_keys is None:
            ancestor_keys = ()
        for key, value in self.items():
            if isinstance(value, TreeNode) and value.is_nested:
                continue  # skip nested values
            pkey = (*ancestor_keys, key)
            base_dict[pkey] = value
        branches = defaultdict(list)
        base_index = tuple(sorted(base_dict.keys()))
        branches[base_index].append(base_dict)
        for key, values in self.items():
            if not isinstance(values, TreeNode) or not values.is_nested:
                continue  # skip non-nested values
            pkey = (*ancestor_keys, key)
            for value, nested in values.items():
                base_dict_copy = copy.deepcopy(base_dict)
                base_dict_copy[pkey] = value
                if not isinstance(nested, TreeNode):
                    msg = f"expected 'TreeNode', got '{type(nested).__name__}'"
                    raise ValueError(msg)
                for branch_index, nested_branches in nested._flatten(
                    base_dict=base_dict_copy,
                    ancestor_keys=pkey,
                ).items():
                    branches[branch_index] += nested_branches
        if len(branches) > 1:
            branches.pop(base_index)
        return branches

    def flatten(self) -> Dict[Tuple, List]:
        return self._flatten()



from __future__ import annotations

import functools
import inspect
from dataclasses import dataclass, is_dataclass
from typing import (
    Any,
    MutableMapping,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from omegaconf import OmegaConf
from typing_extensions import Self

__all__ = [
    "Config",
]


T = TypeVar("T", bound=dataclass)


class ConfigWrapper:
    def __init__(self, wrapped, config: Config, name: Optional[str] = None):
        if name is None:
            name = getattr(wrapped, "__name__", None)
        if name is None:
            name = getattr(wrapped, "__qualname__", None)
        if name is None:
            msg = "builder must have a name"
            raise ValueError(msg)
        self.wrapped = wrapped
        self.signature = inspect.signature(wrapped)
        filter_keys = set()
        for param in self.signature.parameters.values():
            if param.kind == param.POSITIONAL_OR_KEYWORD:
                filter_keys.add(param.name)
        self.filter_keys = filter_keys
        self.name = name
        self.config = config

    def _update_params_from_config(self, args: dict) -> dict:
        config = self.config
        for part in self.name.split("."):
            if part not in config:
                # function or class name or alias is not in config
                return args
            config = config[part]
        # update args
        for k in config:
            if k in args:
                # and args[k] is not __placeholder__
                # already set - do not override
                continue
            args[k] = config[k]
        return args

    def get_params(self, *args, **kwargs):
        args = self.signature.bind_partial(*args, **kwargs)
        args = args.arguments
        args = self._update_params_from_config(args)
        args = {
            k: v
            for k, v in args.items()
            if k in self.filter_keys and v is not inspect.Parameter.empty
        }
        args = self.signature.bind_partial(**args)
        args.apply_defaults()
        return args.arguments

    def __call__(self, *args, **kwargs):
        return self.wrapped(**self.get_params(*args, **kwargs))


def configmethod(func: T) -> T:
    @functools.wraps(func)
    def decoratior(cls, *args, **kwargs):
        return cls(func(cls, *args, **kwargs))

    return decoratior


class Config(MutableMapping):
    @overload
    def __new__(cls, config: Type[T]) -> T: ...

    @overload
    def __new__(cls, config: dict[str, Any]) -> Self: ...

    def __new__(cls, config: Union[Type[T], dict[str, Any]]) -> Union[T, Self]:
        return super().__new__(cls)

    def __init__(self, config: Any) -> None:
        self.omconf = (
            config
            if isinstance(config, OmegaConf)
            else OmegaConf.structured(config)
            if is_dataclass(config)
            else OmegaConf.create(config)
        )

    def __getitem__(self, name: str) -> Any:
        return self.omconf[name]

    def __setitem__(self, name: str, value: Any) -> None:
        self.omconf[name] = value

    def __delitem__(self, name: str) -> None:
        del self.omconf[name]

    def __iter__(self) -> Any:
        return iter(self.omconf)

    def __len__(self) -> int:
        return len(self.omconf)

    def __getattr__(self, name: str) -> Any:
        return self.omconf[name]

    @overload
    def wraps(
        self, wrapped: Union[type, callable], **kwargs
    ) -> ConfigWrapper: ...

    @overload
    def wraps(
        self, wrapped: Union[str, None], **kwargs
    ) -> functools.partial: ...

    def wraps(
        self, wrapped: Union[type, callable, str, None] = None, **kwargs: Any
    ) -> Union[ConfigWrapper, functools.partial]:
        if isinstance(wrapped, str):
            return functools.partial(self.wraps, name=wrapped)
        elif wrapped is None:
            # empty builder call
            return functools.partial(self.wraps, **kwargs)
        config_wrapper = ConfigWrapper(
            wrapped, self, name=kwargs.get("name", None)
        )
        if isinstance(config_wrapper, type):  # if class
            orig_init = config_wrapper.__init__

            @functools.wraps(config_wrapper.__init__)
            def __init__(self, *args, **kwargs):  # noqa: N807
                orig_init(self, **config_wrapper.get_params(*args, **kwargs))

            config_wrapper.__init__ = __init__
            return config_wrapper

        @functools.wraps(config_wrapper)
        def builder_wrapper(*args, **kwargs):
            return config_wrapper(*args, **kwargs)

        return builder_wrapper

    @classmethod
    @configmethod
    def load(cls, path: str) -> Self:
        return OmegaConf.load(path)

    @classmethod
    @configmethod
    def from_dotlist(cls, dotlist: str) -> Self:
        return OmegaConf.from_dotlist(dotlist)

    @classmethod
    @configmethod
    def from_cli(cls, args: list[str]) -> Self:
        return OmegaConf.from_cli(args)



from __future__ import annotations

import datetime as dt
from collections import deque
from contextlib import contextmanager
from typing import ClassVar, Dict, List, Optional, Tuple, Union
from urllib.parse import ParseResult, urlparse

import sqlalchemy
from filelock import FileLock
from sqlalchemy import (
    JSON,
    URL,
    Boolean,
    Column,
    ColumnExpressionArgument,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    String,
    Table,
    case,
    create_engine,
    desc,
)
from sqlalchemy.orm import Session, registry
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound

from octoflow.tracking.models import (
    Experiment,
    JSONType,
    Run,
    RunTags,
    Tag,
    Value,
    Variable,
)
from octoflow.tracking.store import TrackingStore, ValueType, VariableType

__all__ = [
    "SQLAlchemyTrackingStore",
]


mapper_registry = registry()


class SQLAlchemyModelMixin:
    @sqlalchemy.orm.reconstructor
    def init_on_load(self):
        # when object is constructed via sqlalchemy.orm
        # get a ref to the session maker
        self.__post_init__()


class Experiment(Experiment, SQLAlchemyModelMixin, registry=mapper_registry):
    __table__: ClassVar[Table] = Table(
        "experiment",
        mapper_registry.metadata,
        Column("id", Integer, primary_key=True, autoincrement=True),
        Column("name", String(60), nullable=False, unique=True),
        Column("description", String(60), nullable=True),
        Column("artifact_uri", String(60), nullable=True),
    )


class Run(Run, SQLAlchemyModelMixin, registry=mapper_registry):
    __table__: ClassVar[Table] = Table(
        "run",
        mapper_registry.metadata,
        Column("id", Integer, primary_key=True, autoincrement=True),
        Column(
            "experiment_id",
            Integer,
            ForeignKey("experiment.id", ondelete="CASCADE"),
            nullable=False,
        ),
        Column("name", String(60), nullable=False),
        Column("description", String(60), nullable=True),
        Column(
            "created_at", DateTime, nullable=False, default=dt.datetime.utcnow
        ),
    )


class Value(Value, SQLAlchemyModelMixin, registry=mapper_registry):
    __table__: ClassVar[Table] = Table(
        "value",
        mapper_registry.metadata,
        Column("id", Integer, primary_key=True, autoincrement=True),
        Column(
            "run_id",
            Integer,
            ForeignKey("run.id", ondelete="CASCADE"),
            nullable=False,
        ),
        Column(
            "variable_id",
            Integer,
            ForeignKey("variable.id", ondelete="CASCADE"),
            nullable=True,
        ),
        Column("value", JSON, nullable=True),
        Column(
            "timestamp", DateTime, nullable=False, default=dt.datetime.utcnow
        ),
        Column(
            "step_id",
            Integer,
            ForeignKey("value.id", ondelete="CASCADE"),
            nullable=True,
        ),
    )


class Variable(Variable, SQLAlchemyModelMixin, registry=mapper_registry):
    __table__: ClassVar[Table] = Table(
        "variable",
        mapper_registry.metadata,
        Column("id", Integer, primary_key=True, autoincrement=True),
        Column(
            "experiment_id",
            Integer,
            ForeignKey("experiment.id", ondelete="CASCADE"),
            nullable=False,
        ),
        Column("key", String(60), nullable=False),
        Column(
            "parent_id",
            Integer,
            ForeignKey("variable.id", ondelete="CASCADE"),
            nullable=True,
        ),
        Column("type", String(60), nullable=True),
        Column("is_step", Boolean, nullable=True, default=None),
    )


variable_constraints = (
    Index(
        "ix_experiment_id_key",
        Variable.experiment_id,
        Variable.key,
        case(
            (Variable.parent_id.is_(None), "<NULL>"),
            else_=Variable.parent_id,
        ),
        unique=True,
    ),
)


class Tag(Tag, SQLAlchemyModelMixin, registry=mapper_registry):
    __table__: ClassVar[Table] = Table(
        "tag",
        mapper_registry.metadata,
        Column("id", Integer, primary_key=True, autoincrement=True),
        Column("name", String(60), nullable=False, unique=True),
    )


class RunTags(RunTags, SQLAlchemyModelMixin, registry=mapper_registry):
    __table__: ClassVar[Table] = Table(
        "run_tags",
        mapper_registry.metadata,
        Column("id", Integer, primary_key=True, autoincrement=True),
        Column(
            "run_id",
            Integer,
            ForeignKey("run.id", ondelete="CASCADE"),
            nullable=False,
        ),
        Column(
            "tag_id",
            Integer,
            ForeignKey("tag.id", ondelete="CASCADE"),
            nullable=False,
        ),
        Column("value", JSON, nullable=True, default=None),
        Index("ix_run_id_tag_id", "run_id", "tag_id", unique=True),
    )


class SQLAlchemyTrackingStore(TrackingStore):
    """SQLAlchemy tracking store.

    This class is used to define the interface for tracking store.
    """

    def __init__(
        self,
        url: Union[str, URL] = "sqlite:///:memory:",
    ):
        parsed_url: ParseResult = urlparse(url)
        lockfile = None
        if parsed_url.scheme in {"sqlite+filelock", "sqlite+lock"}:
            if len(parsed_url.netloc) > 0:
                msg = "SQLite database with filelock does not support netloc"
                raise ValueError(msg)
            if parsed_url.path == "/:memory:":
                msg = (
                    "SQLite in-memory database is not supported with filelock"
                )
                raise ValueError(msg)
            lockfile = parsed_url.path + ".lock"
            url = f"sqlite:///{parsed_url.path}"
        if lockfile is not None:
            lockfile = FileLock(lockfile)
        self.lock: Optional[FileLock] = lockfile
        self.engine = create_engine(url)
        self.create_all()

    def create_all(self, checkfirst: bool = True):
        mapper_registry.metadata.create_all(
            self.engine,
            checkfirst=checkfirst,
        )

    @contextmanager
    def session(self):
        if self.lock is not None:
            try:
                self.lock.acquire(timeout=10)
            except TimeoutError as e:
                # another process might be using the lock
                raise e
        try:
            with Session(self.engine) as session:
                try:
                    yield session
                except Exception as e:
                    raise e
                finally:
                    deque(map(session.refresh, iter(session)))
                    session.expunge_all()
        finally:
            if self.lock is not None:
                self.lock.release()

    def create_experiment(
        self,
        name: str,
        description: Optional[str] = None,
        artifact_uri: Optional[str] = None,
    ) -> Experiment:
        expr = Experiment(
            name=name,
            description=description,
            artifact_uri=artifact_uri,
        )
        with self.session() as session:
            try:
                session.add(expr)
                session.commit()
            except Exception as e:
                session.rollback()
                msg = f"experiment with name '{name}' already exists"
                raise ValueError(msg) from e
        return expr

    def list_experiments(self) -> List[Experiment]:
        with self.session() as session:
            exprs = session.query(Experiment).all()
        return exprs

    def get_experiment(self, experiment_id: int) -> Experiment:
        with self.session() as session:
            expr = session.query(Experiment).get(experiment_id)
        if expr is None:
            msg = f"experiment with id '{experiment_id}' does not exist"
            raise ValueError(msg)
        return expr

    def get_experiment_by_name(self, name: str) -> Experiment:
        with self.session() as session:
            try:
                stmt = session.query(Experiment).filter(
                    Experiment.name == name
                )
                expr = stmt.one()
            except NoResultFound as e:
                msg = f"experiment with name '{name}' does not exist"
                raise ValueError(msg) from e
            except MultipleResultsFound as e:
                msg = f"multiple experiments with name '{name}' found"
                raise ValueError(msg) from e
        return expr

    def create_run(
        self, experiment_id: int, name: str, description: Optional[str] = None
    ) -> Run:
        run = Run(
            experiment_id=experiment_id,
            name=name,
            description=description,
        )
        with self.session() as session:
            try:
                session.add(run)
                session.commit()
            except Exception as e:
                session.rollback()
                msg = f"could not create run with name '{name}'"
                raise ValueError(msg) from e
        return run

    def delete_run(self, experiment_id: int, run_id: int) -> None:
        with self.session() as session:
            stmt = session.query(Run).filter(
                Run.experiment_id == experiment_id, Run.id == run_id
            )
            run = stmt.one_or_none()
            if run is None:
                msg = f"run with id '{run_id}' does not exist"
                raise ValueError(msg)
            try:
                session.delete(run)
                session.commit()
            except Exception as e:
                session.rollback()
                msg = f"could not delete run with id '{run_id}'"
                raise ValueError(msg) from e

    def search_runs(
        self,
        experiment_id: int,
        expression: Optional[ColumnExpressionArgument[bool]] = None,
    ) -> List[Run]:
        with self.session() as session:
            stmt = session.query(Run).filter(
                Run.experiment_id == experiment_id
            )
            if expression is not None:
                stmt = stmt.filter(expression)
            runs = stmt.order_by(desc(Run.id)).all()
        return runs

    def list_runs(self, experiment_id: int) -> List[Run]:
        return self.list_runs(experiment_id)

    def set_tag(
        self, run_id: int, name: str, value: JSONType = None
    ) -> RunTags:
        msg = f"could not set tag '{name}' for run with id '{run_id}'"
        with self.session() as session:
            tag = Tag(name=name)
            try:
                session.add(tag)
                session.commit()
            except Exception:
                session.rollback()
                tag = session.query(Tag).filter(Tag.name == name).one_or_none()
            if tag is None:
                # unale to get or create tag
                raise ValueError(msg)
            run_tag = RunTags(
                run_id=run_id,
                tag_id=tag.id,
                value=value,
            )
            try:
                session.add(run_tag)
                session.commit()
                return run_tag
            except Exception:
                session.rollback()
            run_tag = (
                session.query(RunTags)
                .filter(
                    RunTags.run_id == run_id,
                    RunTags.tag_id == tag.id,
                )
                .first()
            )
            if run_tag is None:
                raise ValueError(msg)
            # try to update the value
            run_tag.value = value
            try:
                session.commit()
            except Exception as e:
                session.rollback()
                raise ValueError(msg) from e
        return run_tag

    def get_tag(self, run_id: int, name: str) -> JSONType:
        with self.session() as session:
            stmt = (
                session.query(RunTags)
                .select_from(RunTags)
                .join(Tag, RunTags.tag_id == Tag.id)
                .filter(
                    RunTags.run_id == run_id,
                    Tag.name == name,
                )
            )
            run_tag = stmt.one_or_none()
        if run_tag is None:
            return None
        return run_tag.value

    def get_tags(self, run_id: int) -> Dict[str, JSONType]:
        with self.session() as session:
            stmt = (
                session.query(Tag, RunTags)
                .select_from(RunTags)
                .join(Tag, RunTags.tag_id == Tag.id)
                .filter(RunTags.run_id == run_id)
            )
            results: List[Tuple[Tag, RunTags]] = stmt.all()
        tags = []
        for row in results:
            tags.append((row[0].name, row[1].value))
        return dict(tags)

    def count_tags(self, run_id: int) -> int:
        with self.session() as session:
            stmt = (
                session.query(Tag, RunTags)
                .select_from(RunTags)
                .join(Tag, RunTags.tag_id == Tag.id)
                .filter(RunTags.run_id == run_id)
            )
            count = stmt.count()
        return count

    def delete_tag(self, run_id: int, name: str) -> RunTags:
        msg = f"could not delete tag '{name}' for run with id '{run_id}'"
        with self.session() as session:
            tag = session.query(Tag).filter(Tag.name == name).one_or_none()
            if tag is None:
                return None
            run_tag = (
                session.query(RunTags)
                .filter(
                    RunTags.run_id == run_id,
                    RunTags.tag_id == tag.id,
                )
                .one_or_none()
            )
            if run_tag is None:
                return None
            try:
                session.delete(run_tag)
                session.commit()
            except Exception as e:
                session.rollback()
                raise ValueError(msg) from e
        return run_tag

    def log_value(
        self,
        run_id: int,
        key: str,
        value: ValueType,
        *,
        step_id: Optional[int] = None,
        type: Optional[VariableType] = None,
        value_id: Optional[int] = None,
        is_step: Optional[bool] = None,
    ) -> Value:
        with self.session() as session:
            run: Run = session.query(Run).get(run_id)
        parent_id = None
        if step_id is not None:
            with self.session() as session:
                # return an instance or None if not found
                step: Value = session.query(Value).get(step_id)
            if step is None:
                msg = f"step with key '{key}' does not exist"
                raise ValueError(msg)
            if step.run_id != run_id:
                msg = f"step with key '{key}' does not belong to run with id '{run_id}'"
                raise ValueError(msg)
            parent_id = step.variable_id
            with self.session() as session:
                parent: Variable = session.query(Variable).get(parent_id)
                if parent.is_step is None:
                    parent.is_step = (
                        True  # update variable to be a step variable
                    )
                    try:
                        session.commit()
                    except Exception as e:
                        session.rollback()
                        raise e
                elif not parent.is_step:
                    msg = f"variable with key '{parent.key}' is not marked as a step variable"
                    raise ValueError(msg)
        variable = Variable(
            experiment_id=run.experiment_id,
            key=key,
            type=type,
            parent_id=parent_id,
            is_step=is_step,
        )
        try:
            with self.session() as session:
                session.add(variable)
                session.commit()
        except Exception as ex:
            try:
                with self.session() as session:
                    variable = (
                        session.query(Variable)
                        .filter(
                            Variable.experiment_id == run.experiment_id,
                            Variable.key == key,
                            Variable.parent_id == parent_id,
                        )
                        .one()
                    )
            except Exception as e:
                raise e from ex
            if type is not None and variable.type != type:
                msg = f"expected type '{variable.type}', got '{type}' for variable with key '{variable.key}'"
                raise ValueError(msg) from ex
            if is_step is not None and variable.is_step is not is_step:
                msg = f"expected is_step '{variable.is_step}', got '{is_step}' for variable with key '{variable.key}'"
                raise ValueError(msg) from ex
        value = Value(
            run_id=run_id,
            variable_id=variable.id,
            value=value,
            step_id=step_id,
        )
        if value_id is not None:
            value.id = value_id
        try:
            with self.session() as session:
                session.add(value)
                session.commit()
        except Exception as e:
            session.rollback()
            raise e
        return value

    def get_values(self, run_id: int) -> List[Tuple[Variable, Value]]:
        with self.session() as session:
            stmt = (
                session.query(Variable, Value)
                .select_from(Value)
                .join(
                    Variable,
                    Value.variable_id == Variable.id,
                )
                .filter(Value.run_id == run_id)
            )
            values = stmt.order_by(Value.id).all()
        return values



from __future__ import annotations

from typing import Any, Union

import pyarrow.dataset as ds
from pyarrow.dataset import scalar as pyarrow_scalar

from octoflow.data.base import BaseExpression

__all__ = [
    "Expression",
]


class Expression(BaseExpression):
    """A class representing an expression in Octoflow."""

    def __init__(self, expression: Union[Expression, ds.Expression]):
        """
        Initialize an expression.

        Parameters
        ----------
        expression : Union[Expression, ds.Expression]
            The (pyarrow) expression to wrap.
        """
        if isinstance(expression, Expression):
            expression = expression._wrapped
        super().__init__(expression)

    def __eq__(self, other: Any) -> Expression:
        """
        Compare two expressions for equality.

        Parameters
        ----------
        other : Any
            The other expression to compare to.

        Returns
        -------
        Expression
            The expression representing the result of the comparison.
        """
        return Expression(self._wrapped == Expression(other)._wrapped)

    def __ne__(self, other: Any) -> Expression:
        """
        Compare two expressions for inequality.

        Parameters
        ----------
        other : Any
            The other expression to compare to.

        Returns
        -------
        Expression
            The expression representing the result of the comparison.
        """
        return Expression(self._wrapped != Expression(other)._wrapped)

    def __lt__(self, other: Any) -> Expression:
        """
        Compare two expressions for less than.

        Parameters
        ----------
        other : Any
            The other expression to compare to.

        Returns
        -------
        Expression
            The expression representing the result of the comparison.
        """
        return Expression(self._wrapped < Expression(other)._wrapped)

    def __le__(self, other: Any) -> Expression:
        """
        Compare two expressions for less than or equal to.

        Parameters
        ----------
        other : Any
            The other expression to compare to.

        Returns
        -------
        Expression
            The expression representing the result of the comparison.
        """
        return Expression(self._wrapped <= Expression(other)._wrapped)

    def __gt__(self, other: Any) -> Expression:
        """
        Compare two expressions for greater than.

        Parameters
        ----------
        other : Any
            The other expression to compare to.

        Returns
        -------
        Expression
            The expression representing the result of the comparison.
        """
        return Expression(self._wrapped > Expression(other)._wrapped)

    def __ge__(self, other: Any) -> Expression:
        """
        Compare two expressions for greater than or equal to.

        Parameters
        ----------
        other : Any
            The other expression to compare to.

        Returns
        -------
        Expression
            The expression representing the result of the comparison.
        """
        return Expression(self._wrapped >= Expression(other)._wrapped)

    def __and__(self, other: Any) -> Expression:
        """
        Combine two expressions with a logical and.

        Parameters
        ----------
        other : Any
            The other expression to combine with.

        Returns
        -------
        Expression
            The expression representing the result of the combination.
        """
        return Expression(self._wrapped & Expression(other)._wrapped)

    def __or__(self, other: Any) -> Expression:
        """
        Combine two expressions with a logical or.

        Parameters
        ----------
        other : Any
            The other expression to combine with.

        Returns
        -------
        Expression
            The expression representing the result of the combination.
        """
        return Expression(self._wrapped | Expression(other)._wrapped)

    def __invert__(self) -> Expression:
        """
        Invert an expression.

        Returns
        -------
        Expression
            The expression representing the inverted expression.
        """
        return Expression(~self._wrapped)

    def is_nan(self) -> Expression:
        """
        Check if an expression is NaN.

        Returns
        -------
        Expression
            The expression representing the result of the check.
        """
        return Expression(self._wrapped.is_nan())

    def is_null(self, nan_is_null: bool = False):
        """
        Check if an expression is null.

        Parameters
        ----------
        nan_is_null : bool, optional
            Whether to consider NaN values as null, by default False

        Returns
        -------
        Expression
            The expression representing the result of the check.
        """
        return Expression(self._wrapped.is_null(nan_is_null=nan_is_null))

    def is_valid(self) -> Expression:
        """
        Check if an expression is valid.

        Returns
        -------
        Expression
            The expression representing the result of the check.
        """
        return Expression(self._wrapped.is_valid())

    def isin(self, other: Expression) -> Expression:
        """
        Check if an expression is in a set of values.

        Parameters
        ----------
        other : Expression
            The set of values to check against.

        Returns
        -------
        Expression
            The expression representing the result of the check.
        """
        return Expression(self._wrapped.isin(Expression(other)._wrapped))

    def equals(self, other: Expression) -> Expression:
        """
        Check if an expression is equal to another expression.

        Parameters
        ----------
        other : Expression
            The other expression to check against.

        Returns
        -------
        Expression
            The expression representing the result of the check.
        """
        return Expression(self._wrapped.equals(Expression(other)._wrapped))

    def __hash__(self) -> int:
        """
        Get the hash of the expression.

        Returns
        -------
        int
            The hash of the expression.
        """
        return hash(self._wrapped)

    def __repr__(self) -> str:
        """
        Get the representation of the expression.

        Returns
        -------
        str
            The representation of the expression.
        """
        return f"Expression({self._wrapped!r})"


def scalar(value: Any) -> Expression:
    """
    Create an expression from a scalar.

    Parameters
    ----------
    value : Any
        The value of the scalar.

    Returns
    -------
    Expression
        The expression representing the scalar.
    """
    return Expression(pyarrow_scalar(value))



from __future__ import annotations

from pathlib import Path
from typing import Optional, Union

from octoflow.data.dataclass import field
from octoflow.data.dataset import DEFAULT_FORMAT, Dataset
from octoflow.data.expression import Expression, scalar
from octoflow.data.loaders import dataloader
from octoflow.data.loaders import loaders as _loaders

__all__ = [
    "Dataset",
    "Expression",
    "dataloader",
    "field",
    "load_dataset",
    "scalar",
]


def load_dataset(
    __loader: str,
    __path: Optional[str],
    __force: bool = False,
    __dataset_format: str = DEFAULT_FORMAT,
    __dataset_path: Union[Path, str, None] = None,
    /,
    *args,
    **kwargs,
) -> Dataset:
    """
    Load a dataset from a path.

    Parameters
    ----------
    __loader : str
        The name of the loader.
    __path : Optional[str]
        The path to the data (to be passed to the loader).
    __dataset_format : str, optional
        The format of the dataset, by default DEFAULT_FORMAT.
    __dataset_path : Union[Path, str, None], optional
        The path that the dataset will be stored.
    *args : tuple
        The arguments to pass to the loader.
    **kwargs : dict
        The keyword arguments to pass to the loader.

    Returns
    -------
    Dataset
        The loaded dataset.
    """
    loader = _loaders.get(__loader)
    if loader is None:
        # notify the user that the loader is not found
        msg = f"loader '{__loader}' not found"
        raise ValueError(msg)
    if loader.path_arg is not None:
        kwargs[loader.path_arg] = __path
    else:
        # assume the first argument is the path
        args = (__path, *args)
    return Dataset(
        loader,
        __dataset_format,
        path=__dataset_path,
        loader_args=args,
        loader_kwargs=kwargs,
        force=__force,
    )



from __future__ import annotations

import functools
import weakref
from collections import defaultdict
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)
from typing import MutableSet as MutableSetType

__all__ = [
    "flatten",
]


__all__ = [
    "MutableCollection",
    "MutableDict",
    "MutableList",
    "MutableSet",
]

K, V = TypeVar("K"), TypeVar("V")


class EventTarget:
    def __init__(self, *args, **kwargs):
        super(EventTarget, self).__init__()
        self._event_listeners: Dict[str, Set[callable]] = defaultdict(set)

    def add_event_listener(self, type: str, listener: callable):
        self._event_listeners[type].add(listener)

    def remove_event_listener(self, type: str, listener: callable):
        self._event_listeners[type].remove(listener)

    def dispatch_event(self, event: str):
        for callback in self._event_listeners[event]:
            callback()


class MutableCollection(EventTarget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parent_ref: MutableCollection = None

    @property
    def _parent(self) -> Optional[MutableCollection]:
        if self._parent_ref is None:
            return None
        return self._parent_ref()

    def set_parent(self, parent: MutableCollection) -> MutableCollection:
        self._parent_ref = weakref.ref(parent)
        return self

    def changed(self):
        if self._parent is not None:
            self._parent.changed()
        self.dispatch_event("change")

    def coerce(self, value: any):
        if isinstance(value, (list, MutableList)):
            return MutableList(value).set_parent(self)
        elif isinstance(value, (dict, MutableDict)):
            return MutableDict(value).set_parent(self)
        elif isinstance(value, (set, MutableSet)):
            return MutableSet(value).set_parent(self)
        return value


class MutableDict(MutableCollection, MutableMapping[K, V]):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._data = {}
        self.update(dict(*args, **kwargs))

    def __getitem__(self, key: K) -> V:
        return self._data[key]

    def __setitem__(self, key: K, value: V):
        self._data[key] = self.coerce(value)
        self.changed()

    def __delitem__(self, key: K):
        del self._data[key]
        self.changed()

    def __iter__(self) -> Generator[K, None, None]:
        yield from self._data

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return repr(self._data)


class MutableList(MutableCollection, MutableSequence[V]):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._data = []
        self.extend(list(*args, **kwargs))

    def __getitem__(self, key: int) -> V:
        return self._data[key]

    def __setitem__(self, key: int, value: V):
        self._data[key] = self.coerce(value)
        self.changed()

    def insert(self, index: int, value: V):
        self._data.insert(index, self.coerce(value))
        self.changed()

    def __delitem__(self, key: int):
        del self._data[key]
        self.changed()

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return repr(self._data)

    @functools.wraps(list.sort)
    def sort(self, *args, **kwargs) -> None:
        self._data.sort(*args, **kwargs)
        self.changed()


class MutableSet(MutableCollection, MutableSetType[V]):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._data = set()
        self.update(set(*args, **kwargs))

    def add(self, item: V):
        self._data.add(self.coerce(item))
        self.changed()

    def update(self, *s: Iterable[V]) -> None:
        self._data.update({self.coerce(i) for i in set(*s)})
        self.changed()

    def discard(self, item: V):
        self._data.discard(item)
        self.changed()

    def __contains__(self, item: V) -> bool:
        return item in self._data

    def __iter__(self) -> Generator[V, None, None]:
        yield from self._data

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return repr(self._data)


def flatten(
    data: Dict[str, Any],
    *,
    separator: str = ".",
    parent_key: Optional[str] = None,
) -> Dict[Union[str, Tuple[str]], Any]:
    """
    Flatten a nested dictionary.

    Parameters
    ----------
    data : Dict[str, Any]
        The nested dictionary to flatten.
    separator : str, optional
        The separator, by default "."
    parent_key : Optional[str], optional
        The parent key, by default None

    Returns
    -------
    Dict[str | tuple[str], Any]
        The flattened dictionary.
    """
    if separator is not None and parent_key is None:
        parent_key = ""
    items = []
    for key, value in data.items():
        # escape dots
        if separator is None:
            new_key = (parent_key, key) if parent_key is not None else key
        else:
            new_key = parent_key + separator + key if parent_key else key
        if isinstance(value, Mapping):
            items.extend(
                flatten(value, parent_key=new_key, separator=separator).items()
            )
        else:
            items.append((new_key, value))
    return dict(items)



from __future__ import annotations

import abc
import json
import shutil
import weakref
from pathlib import Path
from typing import Any, Dict, Generator, List, Type, Union

from octoflow.utils.collections import MutableDict

_handler_types: Dict[str, Type[ArtifactHandler]] = {}


def get_handler_type(name: str) -> Type[ArtifactHandler]:
    try:
        return _handler_types[name]
    except KeyError:
        msg = f"handler '{name}' not found"
        raise ValueError(msg) from None


def get_handler_type_by_object(obj: Any) -> Type[ArtifactHandler]:
    handler = None
    for handler_type in _handler_types.values():
        if not handler_type.can_handle(obj):
            continue
        if handler is not None:
            msg = f"multiple handlers found for '{type(obj).__name__}'"
            raise ValueError(msg)
        handler = handler_type
    if handler is not None:
        return handler
    # if we get here, we didn't find an appropriate handler for the object
    msg = f"'{type(obj).__name__}' has no handler"
    raise ValueError(msg)


def list_handler_types() -> List[str]:
    return list(_handler_types.keys())


class ArtifactMetadata(MutableDict[str, Any]):
    def __init__(self, handler: ArtifactHandler) -> None:
        self.handler_ref = weakref.ref(handler)
        super().__init__(self._load_data())
        self.add_event_listener("change", self._on_change)

    @property
    def handler(self) -> ArtifactHandler:
        return self.handler_ref()

    def _on_change(self) -> Generator[Dict[str, Any], None, None]:
        path = self.handler.path / ".metadata.json"
        with open(path, "w", encoding="utf-8") as f:
            json.dump(self._data, f)

    def _load_data(self) -> Dict[str, Any]:
        path = self.handler.path / ".metadata.json"
        if not path.exists():
            return {}
        with open(path, encoding="utf-8") as f:
            return json.load(f)


class ArtifactHandlerType(abc.ABCMeta):
    def __new__(cls, *args, **kwargs):
        handler_cls = super().__new__(cls, *args)
        base = args[1]  # base class
        if len(base) > 0:
            handler_cls._handler_type_name = kwargs.get("name", args[0])
            _handler_types[handler_cls._handler_type_name] = handler_cls
        return handler_cls

    @property
    def name(cls) -> str:
        return cls._handler_type_name


class ArtifactHandler(metaclass=ArtifactHandlerType):
    def __init__(self, path: Union[str, Path]) -> None:
        """
        Abstract base class for artifact handlers.

        Parameters
        ----------
        path : str
            The path to the artifact
        """
        super().__init__()
        self.path: Path = Path(path)
        if not self.path.exists():
            self.path.mkdir(parents=True)
        self.metadata = ArtifactMetadata(self)

    @abc.abstractmethod
    def load(self) -> Any:
        """
        Load the artifact from the path.

        Returns
        -------
        Any
            The loaded artifact.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def save(self, obj: Any, *args, **kwargs):
        """
        Save the given artifact to the path.

        Parameters
        ----------
        obj : Any
            The artifact to save.
        args : tuple
            Additional positional arguments.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        None
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def can_handle(cls, obj: object) -> bool:
        """
        Return True if this handler can handle the given object type.

        Parameters
        ----------
        obj : object
            The object to check.

        Returns
        -------
        bool
            True if this handler can handle the given object type.
        """
        raise NotImplementedError

    def exists(self) -> bool:
        """
        Return True if the artifact exists.

        Returns
        -------
        bool
            True if the artifact exists.
        """
        raise NotImplementedError

    def unlink(self):
        """
        Unlink/delete the artifact.

        Returns
        -------
        None
            None
        """
        path = Path(self.path)
        if not path.exists():
            return
        if path.is_dir():
            shutil.rmtree(path)
        else:
            path.unlink()



from __future__ import annotations

import functools
import glob
import json
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    TypeVar,
    Union,
    overload,
)

import pandas as pd
from typing_extensions import ParamSpec

from octoflow import logging
from octoflow.data.base import BaseDatasetLoader
from octoflow.utils import functions

logger = logging.get_logger(__name__)

P = ParamSpec("P")
R = TypeVar("R")
F = TypeVar("F", bound=Callable[..., Any])

loaders: Dict[str, DatasetLoader] = {}


class DatasetLoader(BaseDatasetLoader):
    def __init__(
        self,
        func: Callable[..., Any],
        name: Optional[str] = None,
        extensions: Optional[list[str]] = None,
        path_arg: Optional[str] = None,
        wraps: Optional[Callable[P, R]] = None,
    ):
        """
        Initialize a dataset loader.

        Parameters
        ----------
        func : Callable[..., Any]
            The function to decorate.
        name : Optional[str], optional
            The name of the loader, by default None.
        extensions : Optional[list[str]], optional
            The extensions that the loader supports, by default None.
        path_arg : Optional[str], optional
            The name of the argument that is the path, by default None.
        wraps : Optional[Callable[..., Any]], optional
            The function to wrap, by default None.
        """
        super().__init__()
        self.func = func
        if extensions is None:
            extensions = []
        elif isinstance(extensions, str):
            extensions = [extensions]
        self.name = name or self.func.__name__
        self.extensions = extensions
        self.path_arg = path_arg
        self.wraps = wraps

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """
        Call the loader function.

        Parameters
        ----------
        args : tuple
            The arguments to pass to the function.
        kwargs : dict
            The keyword arguments to pass to the function.

        Returns
        -------
        R
            The result of the function.
        """
        return self.func(*args, **kwargs)

    def bind(self, *args: P.args, **kwargs: P.kwargs) -> Callable[..., R]:
        """
        Bind arguments to the loader function.

        Notes
        -----
        This method is useful for creating a partial function with pre-filled
        arguments and keyword arguments. This helps to improve the uniqueness
        of the fingerprint of the dataset.

        Parameters
        ----------
        args : tuple
            The arguments to pre-fill.
        kwargs : dict
            The keyword arguments to pre-fill.

        Returns
        -------
        Callable[..., R]
            The partial function.
        """
        return functions.bind(self, *args, **kwargs)


@overload
def dataloader(
    func: F,
    name: Optional[str] = None,
    extensions: Optional[list[str]] = None,
    wraps: Optional[Callable[P, R]] = None,
    path_arg: Optional[str] = None,
) -> F:
    """
    Decorator to register a function as a dataset loader.

    Parameters
    ----------
    func : Callable[..., Any]
        The function to decorate.
    name : Optional[str], optional
        The name of the loader, by default None.
    extensions : Optional[list[str]], optional
        The extensions that the loader supports, by default None.
    wraps : Optional[Callable[..., Any]], optional
        The function to wrap, by default None.
    path_arg : Optional[str], optional
        The name of the argument that is the path, by default None.

    Returns
    -------
    F
        The input function.
    """
    ...


@overload
def dataloader(
    name: str,
    extensions: Optional[list[str]] = None,
    wraps: Optional[Callable[P, R]] = None,
    path_arg: Optional[str] = None,
) -> Callable[[F], F]:
    """
    Decorator to register a function as a dataset loader.

    Parameters
    ----------
    name : str
        The name of the loader.
    extensions : Optional[list[str]], optional
        The extensions that the loader supports, by default None.
    wraps : Optional[Callable[..., Any]], optional
        The function to wrap, by default None.
    path_arg : Optional[str], optional
        The name of the argument that is the path, by default None.

    Returns
    -------
    Callable[[F], F]
        The decorator.
    """
    ...


def dataloader(
    func: Union[F, str, None] = None,
    name: Optional[str] = None,
    extensions: Optional[list[str]] = None,
    wraps: Optional[Callable[..., Any]] = None,
    path_arg: Optional[str] = None,
) -> Union[F, Callable[[F], F]]:
    """
    Decorator to register a function as a dataset loader.

    Parameters
    ----------
    func : Union[Callable[..., Any], str, None], optional
        The function to decorate, by default None.
    name : Optional[str], optional
        The name of the loader, by default None.
    extensions : Optional[list[str]], optional
        The extensions that the loader supports, by default None.
    wraps : Optional[Callable[..., Any]], optional
        The function to wrap, by default None.
    path_arg : Optional[str], optional
        The name of the argument that is the path, by default None.

    Returns
    -------
    DatasetLoader
        The dataset loader.
    """
    if func is None:
        # out type: Callable[[F], F]
        return functools.partial(
            dataloader,
            name=name,
            extensions=extensions,
            wraps=wraps,
            path_arg=path_arg,
        )
    elif isinstance(func, str):
        # out type: Callable[[F], F]
        return functools.partial(
            dataloader,
            name=func,
            extensions=extensions,
            wraps=wraps,
            path_arg=path_arg,
        )
    loader = DatasetLoader(
        func,
        name=name,
        extensions=extensions,
        path_arg=path_arg,
        wraps=wraps,
    )
    if loader.name in loaders:
        msg = f"loader with name '{loader.name}' already exists"
        raise ValueError(msg)
    loaders[loader.name] = loader
    # out type: F
    return func


@dataloader(name="json", extensions=[".json"], path_arg="path")
def load_json(
    path: Union[str, Path], encoding: str = "utf-8"
) -> Generator[List[Dict], None, None]:
    """
    Load a dataset from a JSON file.

    Parameters
    ----------
    path : str, Path
        The path to the file.
    encoding : str, optional
        The encoding of the file, by default "utf-8".

    Returns
    -------
    dict
        The loaded dataset.
    """
    if isinstance(path, str):
        path = Path(path)
    if path.is_dir():
        logger.info("loading all .json files in the directory '%s'", path)
        path /= "*.json"
    for p in glob.iglob(str(path)):
        with open(p, encoding=encoding) as f:
            yield json.load(f)


@dataloader(
    name="jsonl",
    extensions=[".jsonl", ".ndjson"],
    path_arg="path",
)
def load_jsonl(
    path: Union[str, Path], encoding: str = "utf-8"
) -> Generator[List[Dict], None, None]:
    """
    Load a dataset from a JSONL file.

    Parameters
    ----------
    path : str, Path
        The path to the file.
    encoding : str, optional
        The encoding of the file, by default "utf-8".

    Returns
    -------
    list[dict]
        The loaded dataset.
    """
    if isinstance(path, str):
        path = Path(path)
    logger.debug(
        "Is the provided path '%s' a directory? %s", path, path.is_dir()
    )
    if path.is_dir():
        logger.info("Loading all .jsonl files in the directory '%s'", path)
        path /= "*.jsonl"
    for p in glob.iglob(str(path)):
        with open(p, encoding=encoding) as f:
            yield [json.loads(line) for line in f]


@dataloader(
    name="csv",
    extensions=[".csv", ".tsv"],
    path_arg="path",
)
def load_csv(
    path: Union[str, Path],
    encoding: str = "utf-8",
) -> Generator[List[Dict], None, None]:
    """
    Load a dataset from a CSV/TSV file.

    Parameters
    ----------
    path : str, Path
        The path to the file.
    encoding : str, optional
        The encoding of the file, by default "utf-8".

    Returns
    -------
    list[dict]
        The loaded dataset.
    """
    if isinstance(path, str):
        path = Path(path)
    if path.is_dir():
        logger.info("loading all .csv files in the directory '%s'", path)
        path /= "*.csv"
    for p in glob.iglob(str(path)):
        p = Path(p)
        if p.suffix == ".tsv":
            return pd.read_csv(p, sep="\t", encoding=encoding).to_dict(
                orient="records"
            )
        yield pd.read_csv(p, encoding=encoding)



from typing import Generic, TypeVar

__all__ = [
    "Property",
]

T = TypeVar("T")


class Property(Generic[T]):
    def __get__(self, obj, objtype=None) -> T: ...

    def __set__(self, obj, value: T) -> None: ...

    def __delete__(self, obj): ...



import importlib


class Package:
    def __init__(self, name: str, modules: list):
        """
        Initialize a package.

        Parameters
        ----------
        name : str
            Name of the package.
        modules : list
            List of modules to expose. Each module can be a `string` or a `dict`.

            If a `string`, it is the name of the module to import.

            If a `dict`, with keys:
            `name` indicating name of the module to import.
            `package` indicating the name of the package to import the module from.
            If not specified, the module is imported from the current package.
        """  # noqa: E501
        self.name = name
        self.modules = modules

    def import_modules(self):
        """
        Import all modules in the package.

        The modules are imported in the order they are defined in the package.

        Returns
        -------
        None
            The modules are imported silently. If an error occurs, it is raised.

        Raises
        ------
        TypeError
            If a module is not a `string` or a `dict`.
        ImportError
            If a module cannot be imported.
        """
        for module in self.modules:
            if isinstance(module, str):
                module = {"name": module}
            if not isinstance(module, dict):
                msg = f"expected str or dict, got {type(module).__name__}"
                raise TypeError(msg)
            name, package = module["name"], module.get("package")
            if package is None:
                package = __package__
            try:
                module = importlib.import_module(
                    name=name,
                    package=package,
                )
            except Exception:
                msg = (
                    f"failed to import '{name}' from "
                    f"'{package}' in '{self.name}'"
                )
                raise ImportError(msg) from None
            if not hasattr(module, "__all__"):
                continue
            for name in module.__all__:
                subpackage = getattr(module, name)
                if not isinstance(subpackage, Package):
                    continue
                subpackage.import_modules()



from __future__ import annotations

import dataclasses as dc
import functools
import inspect
from typing import (
    Any,
    Generic,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
)

import pyarrow.dataset as ds
from sqlalchemy import Table
from sqlalchemy.orm import registry as registry_cls
from typing_extensions import dataclass_transform, get_type_hints

from octoflow.data.expression import Expression

__all__ = [
    "BaseModel",
    "ModelMeta",
]

T = TypeVar("T")


class Field(dc.Field, Expression):
    def __init__(
        self,
        name: Optional[str] = None,
        *,
        default=dc.MISSING,
        default_factory=dc.MISSING,
        init=True,
        repr=True,
        hash=None,
        compare=True,
        metadata=None,
        kw_only=dc.MISSING,
    ):
        """Initialize a field getter."""
        if default is not dc.MISSING and default_factory is not dc.MISSING:
            msg = "cannot specify both default and default_factory"
            raise ValueError(msg)
        kwargs = {}
        if inspect.signature(dc.Field).parameters.get("kw_only") is not None:
            # if python >= 3.10 => should be
            # explicitly passed
            kwargs["kw_only"] = kw_only
        super().__init__(
            default=default,
            default_factory=default_factory,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=metadata,
            **kwargs,
        )
        self.name = name
        # will be initialized when accessed
        Expression.__init__(self, ...)

    @property
    def _wrapped(self) -> ds.Expression:
        """Get the wrapped field."""
        if self._wrapped_ is ...:
            self._wrapped_ = ds.field(self.name)
        return self._wrapped_

    @_wrapped.setter
    def _wrapped(self, value):
        """Get the name of the field."""
        self._wrapped_ = value

    def __call__(self, data: Mapping[str, Any]) -> Any:
        """Get the value of the field.

        Parameters
        ----------
        data : dict
            The data to be accessed.
        """
        return data[self.name]


@functools.wraps(dc.field)
def field(*args, **kwargs) -> Field:
    """Create a new field getter."""
    return Field(*args, **kwargs)


@functools.wraps(dc.field)
def field_from_dataclass_field(field: dc.Field) -> Field:
    """Create a new field getter."""
    kwargs = {}
    if hasattr(field, "kw_only"):
        # if python >= 3.10 => should be
        kwargs["kw_only"] = field.kw_only
    return Field(
        name=field.name,
        default=field.default,
        default_factory=field.default_factory,
        init=field.init,
        repr=field.repr,
        hash=field.hash,
        compare=field.compare,
        metadata=field.metadata,
        **kwargs,
    )


class FieldAccessor(tuple, Generic[T]):
    def __new__(cls, obj: Type[T]) -> FieldAccessor[T]:
        values = []
        field_idx_map = {}
        for idx, field in enumerate(dc.fields(obj)):
            values.append(field)
            field_idx_map[field.name] = idx
        self = super().__new__(cls, tuple(values))
        self._field_idx_map = field_idx_map
        return self

    def __getattr__(self, name: str) -> Field:
        return self[self._field_idx_map[name]]


def fields(
    cls: Type[T],
) -> Union[FieldAccessor[T], Type[T]]:
    return FieldAccessor(cls)


@dataclass_transform(field_specifiers=(Field, field))
class ModelMeta(type):
    def __new__(mcs, name, bases, attrs, **kwargs):  # noqa: N804
        # create empty field if not defined
        if "__annotations__" in attrs:
            annotations: dict = attrs["__annotations__"]
            for attr, _ in annotations.items():
                if attr not in attrs:
                    # create empty field - need so that dataclass
                    # will not create it's default field
                    attrs[attr] = field()
                elif isinstance(attrs[attr], Field):
                    # do not need further processing
                    attrs[attr] = attrs[attr]
                elif isinstance(attrs[attr], dc.Field):
                    # convert dataclass-field to field
                    attrs[attr] = field_from_dataclass_field(attrs[attr])
                else:
                    # create field with default value
                    attrs[attr] = field(default=attrs[attr])
        cls = super().__new__(mcs, name, bases, attrs)
        table: Optional[Table] = kwargs.get(
            "table", getattr(cls, "__table__", None)
        )
        if table is not None and not hasattr(cls, "__table__"):
            cls.__table__ = table
        cls = dc.dataclass(cls)
        registry: registry_cls = kwargs.get("registry", None)
        if table is not None and registry is not None:
            return registry.mapped(cls)
        return cls

    def update_forward_refs(cls, **kwargs: Any) -> None:
        # update forward references in all subclasses
        for subclass in cls.__subclasses__():
            subclass.update_forward_refs(**kwargs)
            # update forward references in type hints
        for name, value in get_type_hints(cls, localns=kwargs).items():
            cls.__annotations__[name] = value


class BaseModel(metaclass=ModelMeta):
    def __post_init__(self):
        pass



from octoflow.tracking.models import Experiment, Run, TrackingClient, Value
from octoflow.tracking.sqlalchemy_store import SQLAlchemyTrackingStore
from octoflow.tracking.store import TrackingStore

__all__ = [
    "Experiment",
    "Run",
    "SQLAlchemyTrackingStore",
    "TrackingClient",
    "TrackingStore",
    "Value",
]



import contextlib

from octoflow import logging
from octoflow.config import config
from octoflow.tracking import Experiment, Run, TrackingClient, Value
from octoflow.utils.config import Config

__version__ = "0.0.46"

__all__ = [
    "Config",
    "Experiment",
    "LoggingFactory",
    "Run",
    "TrackingClient",
    "Value",
    "config",
    "logger",
    "logging",
    "server",
]

# create the octoflow root logger
logger = logging.get_logger(
    name=next(iter(__package__.split("."))),  # octoflow logger
    level=config.logging.level,
    handlers="console",
    formatter=config.logging.format,
)

# import default plugins if available
with contextlib.suppress(ImportError):
    from octoflow_plugins import package as default_plugins_package

    default_plugins_package.import_modules()



from __future__ import annotations

import functools
import inspect
from typing import Callable, TypeVar, Union

from typing_extensions import ParamSpec

__all__ = [
    "bind",
]

P = ParamSpec("P")
T = TypeVar("T")


def bind(
    __func: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> Union[functools.partial, Callable[..., T]]:
    args: list = list(args)
    parameters = inspect.signature(__func).parameters
    arg_names = []
    part_args = []
    part_kwargs = {}
    extra_args = []
    for param in parameters.values():
        if param.kind == param.POSITIONAL_ONLY and len(args) > 0:
            arg = args.pop(0)
            part_args.append(arg)
            arg_names.append(param.name)
        elif param.kind == param.VAR_POSITIONAL:
            while len(args) > 0:
                arg = args.pop(0)
                part_args.append(arg)
                arg_names.append(param.name)
        elif param.kind == param.POSITIONAL_OR_KEYWORD:
            if len(args) > 0:
                value = args.pop(0)
                if param.name in kwargs:
                    extra_args.append(value)
                    value = kwargs.pop(param.name)
                part_kwargs[param.name] = value
            else:
                part_kwargs[param.name] = kwargs[param.name]
        elif param.kind == param.KEYWORD_ONLY and param.name in kwargs:
            part_kwargs[param.name] = kwargs.pop(param.name)
        elif param.kind == param.VAR_KEYWORD:
            while len(kwargs) > 0:
                key, value = kwargs.popitem()
                part_kwargs[key] = value
    if len(args) > 0 or len(extra_args) > 0:
        msg = f"too many arguments for function {__func.__name__}"
        raise TypeError(msg)
    return functools.partial(__func, *part_args, **part_kwargs)



from __future__ import annotations

import contextlib
import functools
import os
import shutil
import tempfile
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)

import numpy as np
import pandas as pd
import polars as pl
import pyarrow as pa
from numpy.typing import ArrayLike
from pandas import DataFrame
from pyarrow import dataset as ds
from tqdm import auto as tqdm

from octoflow import logging
from octoflow.data.base import DEFAULT_BATCH_SIZE, DEFAULT_FORMAT, BaseDataset
from octoflow.data.dataclass import BaseModel
from octoflow.data.expression import Expression
from octoflow.data.loaders import DatasetLoader, loaders
from octoflow.data.schema import get_schema, get_schema_from_dataclass
from octoflow.utils import cache, hash

logger = logging.get_logger(__name__)

SourceType = Union[
    str,
    List[str],
    Union[Path, List[Path]],
    "Dataset",
    List["Dataset"],
]


def _map_func_wrapper(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        output = func(*args, **kwargs)
        if isinstance(output, pd.Series):
            return output
        return pd.Series(output)

    return wrapped


class Dataset(BaseDataset):  # noqa: PLR0904
    @overload
    def __init__(
        self,
        data: Union[List[dict], Dict[str, list], DataFrame] = None,
        format: str = DEFAULT_FORMAT,
    ): ...

    @overload
    def __init__(
        self,
        data: Union[List[dict], Dict[str, list], DataFrame],
        format: str = DEFAULT_FORMAT,
        *,
        path: Optional[Union[str, Path]] = None,
        cache_dir: Optional[Union[str, Path]] = None,
    ): ...

    @overload
    def __init__(
        self,
        loader: DatasetLoader,
        format: str = DEFAULT_FORMAT,
        *,
        path: Optional[Union[str, Path]] = None,
        cache_dir: Optional[Union[str, Path]] = None,
        loader_args: Optional[Tuple[Any, ...]] = None,
        loader_kwargs: Optional[Dict[str, Any]] = None,
        force: bool = False,
    ): ...

    def __init__(
        self,
        data_or_loader: Union[
            List[dict], Dict[str, list], DataFrame, DatasetLoader, str
        ] = None,
        format: str = DEFAULT_FORMAT,
        *,
        schema: Union[pa.Schema, BaseModel, None] = None,
        path: Optional[Union[str, Path]] = None,
        cache_dir: Optional[Union[str, Path]] = None,
        loader_args: Optional[Tuple[Any, ...]] = None,
        loader_kwargs: Optional[Dict[str, Any]] = None,
        force: bool = False,
    ):
        """
        Create a new dataset.

        Parameters
        ----------
        data_or_loader : list of dict, dict of list, DataFrame, BaseDatasetLoader, str
            The data to load into the dataset or the (name of) loader to use.
        format : str
            The format of the dataset.
        path : str, Path, None
            The path to the dataset.
        cache_dir : str, Path, None
            The directory to use for caching.
        loader_args : tuple, None
            The arguments to pass to the loader function if provided
                as the first argument.
        loader_kwargs : dict, None
            The keyword arguments to pass to the loader function if provided
                as the first argument.
        """  # noqa: E501
        if isinstance(data_or_loader, str):
            data_or_loader = loaders[data_or_loader]
        if isinstance(data_or_loader, DatasetLoader):
            if loader_args is None:
                loader_args = ()
            if loader_kwargs is None:
                loader_kwargs = {}
            data_or_loader = data_or_loader.bind(*loader_args, **loader_kwargs)
            data = data_or_loader()  # will result in a generator in most cases
        else:
            data = data_or_loader
        if path is None:
            path = generate_unique_path(data_or_loader, cache_dir=cache_dir)
        if force:
            shutil.rmtree(path, ignore_errors=True)
        if format is None:
            format = DEFAULT_FORMAT
        if isinstance(schema, type) and issubclass(schema, BaseModel):
            schema = get_schema_from_dataclass(schema)
        created = write_dataset(path, data, schema=schema, format=format)
        if not created:
            msg = (
                f"existing dataset found at '{path}', loading existing file(s)"
            )
            logger.warning(msg)
        dataset = read_dataset(path, format=format)
        super().__init__(dataset)
        self._path = path
        self._format = format

    @property
    def path(self) -> Path:
        """
        The path to the dataset.

        Returns
        -------
        Path
            The path to the dataset.
        """
        return self._path

    @property
    def format(self) -> str:
        """
        The format of the dataset.

        Returns
        -------
        str
            The format of the dataset.
        """
        return self._format

    @property
    def _wrapped_format_default_extname(self) -> str:
        return self._wrapped.format.default_extname

    def count_rows(self) -> int:
        """
        Count the number of rows in the dataset.

        Returns
        -------
        int
            The number of rows in the dataset.
        """
        return self._wrapped.count_rows()

    def __len__(self) -> int:
        """
        Get the number of rows in the dataset.

        Returns
        -------
        int
            The number of rows in the dataset.
        """
        return self.count_rows()

    def head(
        self,
        num_rows: int = 5,
        columns: Union[str, List[str], None] = None,
        filter: Expression = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> DataFrame:
        """
        Get the first rows of the dataset as a pandas DataFrame.

        Parameters
        ----------
        num_rows : int
            The number of rows to get.
        columns : str, list of str, None
            Names of columns to get. If None, all columns are returned.
        filter : Expression
            The filter expression.
        batch_size : int
            Number of rows to get at a time.

        Returns
        -------
        DataFrame
            A pandas DataFrame containing the first rows of the dataset.
        """
        filter = filter.to_pyarrow() if filter else None
        if isinstance(columns, str):
            columns = [columns]
        table: pa.Table = self._wrapped.head(
            num_rows=num_rows,
            columns=columns,
            filter=filter,
            batch_size=batch_size,
        )
        return table.to_pandas()

    @overload
    def __getitem__(self, indices: int) -> Dict[str, Any]: ...

    @overload
    def __getitem__(
        self, indices: Union[slice, List[int], ArrayLike]
    ) -> pa.Table: ...

    def __getitem__(
        self, indices: Union[int, slice, List[int], ArrayLike]
    ) -> Union[dict, pa.Table]:
        """Get rows from the dataset."""
        return self.take(indices=indices)

    @overload
    def take(
        self,
        *,
        indices: Optional[int] = None,
        columns: Union[str, List[str], None] = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> Dict[str, Any]: ...

    @overload
    def take(
        self,
        *,
        indices: Union[slice, List[int], ArrayLike] = None,
        columns: Union[str, List[str], None] = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> DataFrame: ...

    def take(
        self,
        *,
        indices: Union[int, slice, List[int], ArrayLike] = None,
        columns: Union[str, List[str], None] = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> Union[dict, pa.Table]:
        """
        Take rows from the dataset.

        Parameters
        ----------
        indices : int, slice, list of int, array-like
            Indices of rows to take.
        columns : str, list of str, None
            Names of columns to take. If None, all columns are taken.
        batch_size : int
            Number of rows to take at a time.

        Returns
        -------
        Document, pyarrow.Table
            The taken rows or row.
        """
        return_batch = True
        if indices is None:
            indices = []
        elif isinstance(indices, int):
            return_batch = False
            indices = [indices]
        elif isinstance(indices, slice):
            step_size = indices.step
            if step_size is None:
                # default step size is 1
                step_size = 1
            elif not isinstance(step_size, int):
                msg = "expected indices.step to be int, got '{indices.step}'"
                raise ValueError(msg)
            elif step_size < 1:
                msg = (
                    "expected indices.step to be greater than 0, "
                    f"got '{step_size}'"
                )
                raise ValueError(msg)
            start = indices.start or 0
            stop = indices.stop or self.count_rows()
            indices = np.arange(start, stop, step_size)
        if isinstance(columns, str):
            columns = [columns]
        batch = self._wrapped.take(
            indices=indices,  # array or array-like
            columns=columns,  # list of str or None
            batch_size=batch_size,  # int
        )
        if return_batch:
            return batch
        return batch.to_pylist()[0]

    def map(
        self,
        func: Any,
        batch_size: int = DEFAULT_BATCH_SIZE,
        batched: bool = False,
        verbose: Union[bool, int] = 1,
    ) -> Dataset:
        """
        Map a function over the dataset.

        Parameters
        ----------
        func : Any
            The function to map over the dataset.
        batch_size : int
            Number of rows to map at a time.
        batched : bool
            Whether the function is batched.
        verbose : bool | int
            Whether to show a progress bar.

        Returns
        -------
        Dataset
            A new dataset containing the mapped rows.
        """
        fingerprint = hash.hash(func)
        path = self.path / f"map-{fingerprint}"
        num_batches = ((self.count_rows() - 1) // batch_size) + 1
        batch_iter = self._wrapped.to_batches(batch_size=batch_size)
        if verbose:
            progress_bar = tqdm.tqdm(
                batch_iter,
                total=num_batches,
                desc=f"Mapping [{fingerprint}]",
            )
        else:
            progress_bar = None
        batch_iter = (
            record_batch(
                func(batch)
                if batched
                else batch.to_pandas().apply(
                    _map_func_wrapper(func),
                    axis=1,
                )
            )
            for batch in (batch_iter if progress_bar is None else progress_bar)
        )
        state = write_dataset(
            path,
            batch_iter,
            format=self.format,
        )
        if not state:
            logger.warning(
                f"existing dataset found at '{path}', "
                "loading existing file(s)"
            )
        if progress_bar is not None:
            progress_bar.update(num_batches)
            progress_bar.close()
        return type(self).load_dataset(path, format=self.format)

    def filter(self, expression: Expression = None) -> Dataset:
        """
        Filter the dataset.

        Parameters
        ----------
        expression : Expression
            The filter expression.

        Returns
        -------
        Dataset
            A new dataset containing only the rows that match the filter expression.
        """  # noqa: E501
        if expression is None:
            return self
        pyarrow_expression = expression.to_pyarrow()
        fingerprint = hash.hash(pyarrow_expression)
        path = self.path / f"filter-{fingerprint}"
        dataset = self._wrapped.filter(pyarrow_expression)
        state = write_dataset(
            path,
            dataset,
            schema=dataset.schema,
            format=self.format,
        )
        if not state:
            logger.warning(
                f"existing dataset found at '{path}', "
                "loading existing file(s)"
            )
        return self.load_dataset(path, format=self.format)

    def select(
        self,
        columns: Union[str, List[str]],
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> Dataset:
        """
        Select columns from the dataset.

        Parameters
        ----------
        columns : str, list of str
            Names of columns to select.

        Returns
        -------
        Dataset
            A new dataset containing only the selected columns.
        """
        if isinstance(columns, str):
            columns = [columns]
        fingerprint = hash.hash(columns)
        path = self.path / f"select-{fingerprint}"
        scanner = ds.Scanner.from_dataset(
            self._wrapped,
            columns=columns,
            batch_size=batch_size,
        )
        state = write_dataset(
            path,
            scanner,
            format=self.format,
        )
        if not state:
            logger.warning(
                f"existing dataset found at '{path}', "
                "loading existing file(s)"
            )
        return self.load_dataset(path, format=self.format)

    def cleanup(self):
        """
        Delete the directory containing the dataset.

        Returns
        -------
        None
        """
        if not self.path.exists():
            return
        shutil.rmtree(self.path)

    def __enter__(self) -> Dataset:
        """
        Context manager entry point.

        Returns
        -------
        Dataset
            The dataset.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Context manager exit point.

        This will delete the resources allocated for this dataset.

        Parameters
        ----------
        exc_type : Exception
            The exception type.
        exc_value : Exception
            The exception value.
        traceback : Traceback
            The traceback.

        Returns
        -------
        None
        """
        with contextlib.suppress(Exception):
            self.cleanup()

    @classmethod
    def load_dataset(
        cls,
        path: Union[Path, str],
        format: str = DEFAULT_FORMAT,
    ) -> Dataset:
        """
        Load an existing dataset.

        Parameters
        ----------
        path : str, Path
            The path to the dataset.
        format : str
            The format of the dataset.

        Returns
        -------
        Dataset
            The loaded dataset.
        """
        inst = cls.__new__(cls)
        inst._wrapped = read_dataset(path, format=format)
        inst._path = path
        inst._format = format
        return inst

    def to_polars(self) -> pl.LazyFrame:
        """
        Convert the dataset to a Polars DataFrame.

        Returns
        -------
        pl.LazyFrame
            The Polars Lazy DataFrame.
        """
        return pl.scan_ipc(self.path / "data" / "*.arrow")


def write_dataset(
    path: Union[str, Path],
    data: Union[
        ds.Dataset,
        pa.Table,
        pa.RecordBatch,
        # Iterable[pa.Table],
        Iterable[pa.RecordBatch],  # pa.RecordBatchReader.from_batches
        pa.RecordBatchReader,
        pd.DataFrame,  # from_pandas
        Mapping[str, List[Any]],  # from_pydict
        Sequence[Mapping[str, Any]],  # from_pylist
    ],
    schema: pa.Schema = None,
    format: Optional[str] = None,
) -> bool:
    if isinstance(path, str):
        path = Path(path)
    if not path.exists():
        path.mkdir(
            parents=True,
            exist_ok=True,
        )
    out_data_path = get_data_path(path)
    if out_data_path.exists():
        return False
    # first write to temporary directory
    temp_path = Path(
        tempfile.mkdtemp(
            prefix=".temp-",
            dir=path,
        )
    )
    # it might take some time to write the data
    # within that time, another process might try
    # to read the data or write the data so we
    # write to a temporary directory first
    # and then move the data to the desired
    # directory
    if not isinstance(data, (ds.Dataset, ds.Scanner)):
        data = record_batch(data, schema=schema)
    if isinstance(data, pa.RecordBatchReader):
        schema = None
    ds.write_dataset(data, temp_path, schema=schema, format=format)
    try:
        os.replace(temp_path, out_data_path)
    except OSError:
        return False
    return True


def read_dataset(path: Union[str, Path], format: str) -> ds.dataset:
    data_path = get_data_path(path)
    return ds.dataset(
        data_path,
        format=format,
    )


def get_data_path(path: Union[Path, str]) -> Path:
    if isinstance(path, str):
        path = Path(path)
    if path.is_dir() or not path.exists():
        path.mkdir(parents=True, exist_ok=True)
        data_path = path / "data"
    else:
        msg = f"expected path to be directory, got '{path}'"
        raise ValueError(msg)
    return data_path


def generate_unique_path(
    reference: Any,
    cache_dir: Optional[Union[str, Path]] = None,
) -> Path:
    # create a temporary directory in system temp directory
    if cache_dir is None:
        cache_dir = cache.path / "datasets"
    if cache_dir is not None and not cache_dir.exists():
        try:
            cache_dir.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            msg = f"failed to create cache directory at '{cache_dir}'"
            raise OSError(msg) from e
    try:
        fingerprint = hash.hash(reference)
        path = cache_dir / f"data-{fingerprint}"
    except Exception as e:
        # an error occurred while hashing data
        msg = (
            "Using temporary directory; "
            f"Failed to create unique path due to the following error: {e}."
        )
        logger.warning(msg)
        path = Path(tempfile.mkdtemp(dir=cache_dir))
    return path


def record_batch(
    data: Any, schema: Optional[pa.Schema] = None
) -> Union[pa.RecordBatch, pa.RecordBatchReader]:
    # if generator, return a generator
    if isinstance(data, pa.RecordBatch):
        return data
    if isinstance(data, Generator):
        # Note: generator of batches not instances
        data = (record_batch(d, schema=schema) for d in data)
        if schema is None:
            data, schema = get_schema(data)
        return pa.RecordBatchReader.from_batches(schema, data)
    if isinstance(data, pa.Table):
        # zero-copy
        if schema is None:
            schema = data.schema
        return pa.RecordBatchReader.from_batches(
            schema, (d for d in data.to_batches())
        )
    if isinstance(data, pd.DataFrame):
        return pa.RecordBatch.from_pandas(data, schema=schema)
    if isinstance(data, Mapping):
        return pa.RecordBatch.from_pydict(data, schema=schema)
    if isinstance(data, Sequence) and not isinstance(data, str):
        return pa.RecordBatch.from_pylist(list(data), schema=schema)
    dtype = data.__class__.__name__
    msg = (
        f"expected data to be of type 'pa.RecordBatch', 'pa.Table', "
        f"'pd.DataFrame', Mapping, or Sequence, got '{dtype}'"
    )
    raise ValueError(msg)



import logging
from typing import Optional, Sequence, Union

__all__ = [
    "get_logger",
]


def get_logger(
    name: Optional[str] = None,
    level: Union[int, str, None] = None,
    handlers: Optional[Sequence[str]] = None,
    formatter: Optional[str] = None,
) -> logging.Logger:
    """
    Get a logger with the given name and level.

    Parameters
    ----------
    name : str, optional
        Name of the logger.
    level : int or str, optional
        Logging level.
    formatter : str, optional
        Formatter string.
    handlers : Sequence[str], optional
        Sequence of handlers to use.

    Returns
    -------
    logging.Logger
        Logger instance.
    """
    logger = logging.getLogger(name)
    if level is not None:
        logger.setLevel(level)
    if handlers is None:
        handlers = ()
    elif isinstance(handlers, str):
        handlers = (handlers,)
    if "console" in handlers:
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.DEBUG)
        if not isinstance(formatter, logging.Formatter):
            # Create a formatter
            formatter = logging.Formatter(formatter)
        # add it to the handlers
        stream_handler.setFormatter(formatter)
        # Add the handlers to the logger
        logger.addHandler(stream_handler)
    return logger



from __future__ import annotations

import abc
import datetime as dt
import functools
from contextvars import ContextVar
from types import FunctionType
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    TypedDict,
    Union,
)

from octoflow.data.dataclass import BaseModel

if TYPE_CHECKING:
    from octoflow.tracking.models import (
        Experiment,
        JSONType,
        Run,
        RunTags,
        Value,
        Variable,
    )

__all__ = [
    "TrackingStore",
    "wrap",
]

store_cv = ContextVar("store", default=None)

VariableType = Literal["param", "metric"]

ValueType = Union[str, float, int, bool, None]


class StoredModel(BaseModel):
    def __post_init__(self):
        super().__post_init__()
        store = store_cv.get()
        if store is None:
            msg = "unable to bind object context to a tracking store"
            raise RuntimeError(msg)
        self._store: TrackingStore = store

    @property
    def store(self) -> TrackingStore:
        store = store_cv.get()
        if store is None:
            msg = "tracking store context is not set"
            raise RuntimeError(msg)
        return store


def wrap(method):
    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        if isinstance(self, StoredModel):
            store = self._store
        elif isinstance(self, TrackingStore):
            store = self
        with store:
            return method(self, *args, **kwargs)

    return wrapped


class ValueMapping(TypedDict):
    key: str
    value: ValueType
    type: VariableType
    step_id: Optional[int]
    timestamp: Optional[dt.datetime]
    is_step: Optional[bool]


class ValueTuple(NamedTuple):
    key: str
    value: ValueType
    type: VariableType
    step_id: Optional[int] = None
    timestamp: Optional[dt.datetime] = None
    is_step: Optional[bool] = None


class TrackingStoreMetaClass(abc.ABCMeta):
    def __new__(cls, name, bases, attrs: Dict[str, Any], **kwargs):
        for method_name in attrs:
            method = attrs[method_name]
            if method_name.startswith("_"):
                continue
            if not isinstance(method, FunctionType):
                continue
            attrs[method_name] = wrap(method)
        return super().__new__(cls, name, bases, attrs, **kwargs)


class TrackingStore(metaclass=TrackingStoreMetaClass):
    """Abstract class for tracking store.

    This class is used to define the interface for tracking store.
    """

    def __enter__(self):
        if not hasattr(self, "_tokens") or self._tokens is None:
            self._tokens = []
        self._tokens.append(store_cv.set(self))
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not hasattr(self, "_tokens"):
            return
        store_cv.reset(self._tokens.pop())

    @abc.abstractmethod
    def create_experiment(
        self,
        name: str,
        description: Optional[str] = None,
        artifact_uri: Optional[str] = None,
    ) -> Experiment:
        raise NotImplementedError

    @abc.abstractmethod
    def list_experiments(self) -> List[Experiment]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_experiment(self, experiment_id: int) -> Experiment:
        raise NotImplementedError

    @abc.abstractmethod
    def get_experiment_by_name(self, name: str) -> Experiment:
        raise NotImplementedError

    @abc.abstractmethod
    def create_run(
        self,
        experiment_id: int,
        name: str,
        description: Optional[str] = None,
    ) -> Run:
        raise NotImplementedError

    @abc.abstractmethod
    def delete_run(self, experiment_id: int, run_id: int) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def list_runs(self, experiment_id: int) -> List[Run]:
        raise NotImplementedError

    @abc.abstractmethod
    def search_runs(self, experiment_id: int, **kwargs) -> List[Run]:
        raise NotImplementedError

    @abc.abstractmethod
    def set_tag(
        self, run_id: int, name: str, value: JSONType = None
    ) -> RunTags:
        raise NotImplementedError

    @abc.abstractmethod
    def get_tag(self, run_id: int, name: str) -> JSONType:
        raise NotImplementedError

    @abc.abstractmethod
    def get_tags(self, run_id: int) -> Dict[str, JSONType]:
        raise NotImplementedError

    @abc.abstractmethod
    def count_tags(self, run_id: int) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def delete_tag(self, run_id: int, name: str) -> RunTags:
        raise NotImplementedError

    @abc.abstractmethod
    def log_value(
        self,
        run_id: int,
        key: str,
        value: str,
        *,
        step_id: Optional[int] = None,
        type: Optional[VariableType] = None,
        is_step: Optional[bool] = None,
    ) -> Value:
        raise NotImplementedError

    def _log_value(
        self,
        run_id: int,
        value: Union[ValueMapping, ValueTuple],
        *,
        step_id: Optional[int] = None,
        type: Optional[VariableType] = None,
    ) -> Value:
        if isinstance(value, Mapping):
            value = ValueTuple(**value)
        elif isinstance(value, ValueTuple):
            pass
        elif isinstance(value, Tuple):
            value = ValueTuple(*value)
        else:
            msg = (
                f"expected 'dict' or 'tuple', got '{value.__class__.__name__}'"
            )
            raise TypeError(msg)
        return self.log_value(
            run_id,
            value.key,
            value.value,
            step_id=step_id or value.step_id,
            type=type or value.type,
            is_step=value.is_step,
        )

    def log_values(
        self,
        run_id: int,
        values: List[Union[ValueMapping, ValueTuple, Value]],
        *,
        step_id: Optional[int] = None,
        type: Optional[VariableType] = None,
    ) -> List[Value]:
        return [
            self._log_value(run_id, value, step_id=step_id, type=type)
            for value in values
        ]

    @abc.abstractmethod
    def get_values(self, run_id: int) -> List[Tuple[Variable, Value]]:
        raise NotImplementedError

    def import_store(self, other: TrackingStore):
        for other_experiment in other.list_experiments():
            try:
                this_experiment = self.create_experiment(
                    other_experiment.name,
                    description=other_experiment.description,
                    artifact_uri=other_experiment.artifact_uri,
                )
            except ValueError:
                this_experiment = self.get_experiment_by_name(
                    other_experiment.name
                )
            for other_run in other.list_runs(other_experiment.id):
                this_run = self.create_run(
                    this_experiment.id,
                    other_run.name,
                    description=other_run.description,
                )
                value_map = {}
                for var, other_value in other.get_values(other_run.id):
                    step_id = (
                        None
                        if other_value.step_id is None
                        else value_map[other_value.step_id]
                    )
                    this_value = self.log_value(
                        this_run.id,
                        var.key,
                        other_value.value,
                        step_id=step_id,
                        type=var.type,
                        is_step=var.is_step,
                    )
                    value_map[other_value.id] = this_value.id
        return self
